"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import time as ti
import os
import re
import shutil as sh
import traceback as tb
import glob
import copy
import subprocess as sp
import tarfile
import numpy as np
import pexpect as pe
try:
    from PyQt4.QtCore import QThread
except ImportError:
    from PyQt5.QtCore import QThread
from transphire import transphire_utils as tu
from transphire import transphire_software as tus
from transphire import transphire_motion as tum
from transphire import transphire_ctf as tuc


class ProcessThread(QThread):
    """
    Worker thread

    Inherits from:
    QThread

    Buttons:
    None

    Signals:
    None
    """

    def __init__(
            self,
            shared_dict,
            name,
            content_settings,
            queue_com,
            password,
            settings,
            mount_directory,
            use_threads_set,
            stop,
            parent=None
            ):
        """
        Initialize object variables.

        Arguments:
        shared_dict - Dictionary containing shared information
        name - Name of the process
        content_settings - Settings for this process
        queue_com - Queue communication dictionary
        password - Sudo password
        settings - TranSPHIRE settings
        mount_directory - Directory for mount points
        use_threads_set - set of threads that are used
        stop - Value of the stop variable
        parent - Parent widget (default None)

        Return:
        None
        """
        super(ProcessThread, self).__init__(parent)
        # Variables
        self.stop = stop
        self.password = password
        self.shared_dict = shared_dict
        self.done = False
        self.content_settings = content_settings

        self.name = name
        self.typ = self.content_settings['name']
        self.run_this_thread = bool(self.typ in use_threads_set)

        self.queue_com = queue_com
        self.settings = settings
        self.mount_directory = mount_directory
        self.time_last = None
        self.time_last_error = None
        self.notification_send = None
        self.notification_time = float(self.settings['General']['Time until notification'])

        self.queue = shared_dict['queue'][self.content_settings['name']]
        self.shared_dict_typ = shared_dict['typ'][self.content_settings['name']]
        self.queue_lock = self.shared_dict_typ['queue_lock']

        try:
            self.later = bool(self.settings['Copy'][self.typ] == 'Later')
        except KeyError:
            self.later = False
        try:
            self.user = self.settings['{0}_user'.format(self.typ)]
        except KeyError:
            self.user = None

        self.queue_com['status'].put(['Starting', name, 'yellow'])

    def run(self):
        """
        Run the thread.

        Arguments:
        None

        Return:
        None
        """
        # Current time
        self.time_last = ti.time()
        self.time_last_error = ti.time()
        self.notification_send = False

        # Event loop
        while not self.stop:
            if self.done:
                while not self.stop:
                    self.queue_com['status'].put([
                        'Finished',
                        self.name,
                        'black'
                        ])
                    QThread.sleep(10)
                continue
            else:
                pass

            if not self.run_this_thread:
                while not self.stop:
                    self.queue_com['status'].put([
                        'Skipped {0}'.format(self.queue.qsize()),
                        self.name,
                        'black'
                        ])
                    QThread.sleep(10)
                continue
            else:
                pass

            if self.later:
                while not self.stop:
                    self.queue_com['status'].put([
                        'Later {0}'.format(self.queue.qsize()),
                        self.name,
                        'black'
                        ])
                    QThread.sleep(10)
                continue
            else:
                pass

            if self.check_quota():
                pass
            else:
                self.queue_com['status'].put([
                    'Quota Error {0}'.format(self.queue.qsize()),
                    self.name,
                    'red'
                    ])
                QThread.sleep(10)
                continue

            if self.check_connection():
                pass
            else:
                self.queue_com['status'].put([
                    'Connection Error {0}'.format(self.queue.qsize()),
                    self.name,
                    'red'
                    ])
                QThread.sleep(10)
                continue

            if self.check_full():
                pass
            else:
                self.queue_com['status'].put([
                    'No space Error {0}'.format(self.queue.qsize()),
                    self.name,
                    'red'
                    ])
                QThread.sleep(10)
                continue

            if self.shared_dict_typ['unknown_error']:
                time_diff = ti.time() - self.time_last
                if self.typ == 'Find':
                    self.queue_com['status'].put([
                        'Unknown Error {0:.1f}min'.format(time_diff / 60),
                        self.name,
                        'red'
                        ])
                else:
                    self.queue_com['status'].put([
                        'Unknown Error {0} {1} {2:.1f}min'.format(
                            self.queue.qsize(),
                            self.shared_dict_typ['file_number'],
                            time_diff / 60
                            ),
                        self.name,
                        'red'
                        ])
                i = 0
                while i < 6:
                    QThread.sleep(10)
                    if self.stop:
                        break
                    else:
                        i += 1
                self.shared_dict_typ['unknown_error'] = False
                continue
            else:
                pass

            # Start processes
            if self.typ == 'Find':
                self.start_queue_find()
            elif self.typ == 'Meta':
                self.start_queue_meta()
            else:
                self.start_queue()

        # Print, if stopped
        self.queue_com['status'].put(['STOPPED', self.name, 'red'])
        print(self.name, ': Stopped')

    def check_full(self):
        """
        Check, if the program got an no space error.

        Arguments:
        None

        Return:
        True, if there is a no space error, else False
        """
        copy_threads = [
            'full_work', 'full_backup', 'full_hdd', 'full_transphire'
            ]
        for name in copy_threads:
            if self.shared_dict_typ[name]:
                return False
            else:
                pass
        return True

    def check_quota(self):
        """
        Check, if the computer reached the quota limit.

        Arguments:
        None

        Return:
        True, if the computer did not hit the limit, else False
        """
        scratch_stop = float(self.settings['General']['Scratch quota stop (%)']) / 100
        project_stop = float(self.settings['General']['Project quota stop (%)']) / 100
        if self.typ in ['Copy_work', 'Copy_backup', 'Copy_hdd']:
            return True
        else:
            pass

        for stop, folder in zip(
                [project_stop, scratch_stop],
                ['project_folder', 'scratch_folder']
                ):
            try:
                total_quota = sh.disk_usage(self.settings[folder]).total / 1e12
                used_quota = sh.disk_usage(self.settings[folder]).used / 1e12
            except FileNotFoundError:
                self.stop = True
                message = ''.join([
                    '{0} no longer available!\n'.format(
                        folder.replace('_', ' ')
                        ),
                    '{0} is stopping now!'.format(self.name)
                    ])
                self.queue_com['notification'].put(
                    message
                    )
                self.queue_com['error'].put(
                    message
                    )
                return False
            if used_quota > (total_quota * stop):
                self.stop = True
                message = ''.join([
                    'Less than {0:.1f} Tb free on {1}!'.format(
                        total_quota * (1 - stop),
                        folder.replace('_', ' ')
                        ),
                    '{0} is stopping now!'.format(self.name)
                    ])
                self.queue_com['notification'].put(message)
                self.queue_com['error'].put(message)
                return False
            else:
                pass
        return True

    def check_connection(self):
        """
        Check, if the process got a lost connection error.

        Arguments:
        None

        Return:
        True, if the process did not get an error, else False
        """
        # List of processes and related folder names
        processes = [
            ['lost_input_frames', 'Search path frames'],
            ['lost_input_meta', 'Search path meta'],
            ['lost_work', 'Copy_work_folder'],
            ['lost_backup', 'Copy_backup_folder'],
            ['lost_hdd', 'Copy_hdd_folder']
            ]
        for process, folder in processes:
            if self.shared_dict_typ[process]:
                time_diff = ti.time() - self.time_last
                if folder == 'Search path frames' or \
                        folder == 'Search path meta':
                    output_folder = self.settings['General'][folder]
                else:
                    output_folder = self.settings[folder]

                self.queue_com['status'].put([
                    '{0}: Lost connection {1:.1f}min'.format(
                        self.typ,
                        time_diff / 60
                        ),
                    self.name,
                    'red'
                    ])
                if self.typ == 'Motion' or \
                        self.typ == 'CTF' or \
                        self.typ == 'Compress':
                    # These processes do not have a mount point
                    return False
                elif self.mount_directory in output_folder:
                    if self.check_if_mounted(output_folder):
                        self.shared_dict_typ[process] = False
                        self.queue_com['notification'].put(''.join([
                            '{0} is connected again!'.format(self.name),
                            'If this message is occuring very often, ',
                            'please remount {0} or restart TranSPHIRE'.format(
                                self.name
                                )
                            ]))
                        if ti.time() - self.time_last_error > 1800:
                            self.queue_com['error'].put(
                                '{0} is connected again!'.format(self.typ)
                                )
                            self.time_last_error = ti.time()
                        else:
                            pass
                    else:
                        return False
                else:
                    pass
            else:
                pass

        return True

    @staticmethod
    def check_if_mounted(directory):
        """
        Check, if the mount point is still present.

        Arguments:
        directory - Directory to check.

        Return:
        True, if the mount point is still present, else False
        """
        folder_names = directory.split('/')
        # Return, if the process is not running.
        if folder_names[-1] == 'False':
            return False
        else:
            pass

        for idx in range(1, len(folder_names)+1):
            current_dir = os.path.join(*folder_names[0:idx])
            if not current_dir:
                continue
            elif folder_names[0]:
                pass
            else:
                current_dir = '/{0}'.format(current_dir)

            if os.path.ismount(current_dir):
                return True
            else:
                try:
                    os.listdir(current_dir)
                except PermissionError:
                    return True
                else:
                    pass
        return False

    def start_queue_meta(self):
        """
        Start copying meta files.

        Arguments:
        None

        Return:
        None
        """
        if not self.settings['Copy_software_meta']:
            self.done = True
            return None
        else:
            pass
        folder = self.settings['General']['Search path meta']
        if folder:
            self.queue_com['status'].put([
                'Copy Metadata',
                self.name,
                'green'
                ])
            try:
                self.run_software_meta(directory=folder)
            except FileNotFoundError:
                self.stop = True
                message = ''.join([
                    '{0} no longer available!\n'.format(
                        folder.replace('_', ' ')
                        ),
                    '{0} is stopping now!'.format(self.name)
                    ])
                self.queue_com['notification'].put(
                    message
                    )
                self.queue_com['error'].put(
                    message
                    )
        else:
            print('No search path specified')
        self.done = True

    def start_queue_find(self):
        """
        Start finding files to process.

        Arguments:
        None

        Return:
        None
        """
        time_diff = ti.time() - self.time_last

        self.queue_com['status'].put([
            'Running {0:.1f}min'.format(time_diff / 60),
            self.name,
            'green'
            ])
        try:
            self.run_find()
        except FileNotFoundError:
            self.write_error(
                msg=tb.format_exc(),
                root_name=''
                )
            self.lost_connection(
                typ='lost_work'
                )
        except Exception:
            print('!!! UNKNOWN !!!\n')
            msg = tb.format_exc()
            self.write_error(
                msg=msg,
                root_name=''
                )
            if 'termios.error' not in msg:
                message = ''.join([
                    'Unknown error occured in {0}!'.format(self.name),
                    'The process will continue, but check the error file!',
                    '\n{0}'.format(self.shared_dict_typ['error_file'])
                    ])
                self.queue_com['notification'].put(message)
                if ti.time() - self.time_last_error > 1800:
                    self.queue_com['error'].put(message)
                    self.time_last_error = ti.time()
                else:
                    pass
                self.shared_dict_typ['unknown_error'] = True
            else:
                pass
        else:
            time_diff = ti.time() - self.time_last
            if time_diff / 60 > self.notification_time and \
                    not self.notification_send:
                self.notification_send = True
                message = 'No new files from data collection in the last {0} Minutes!'.format(
                    self.notification_time
                    )
                self.queue_com['notification'].put(message)
                self.queue_com['error'].put(message)
            QThread.sleep(20)

    def start_queue(self):
        """
        Start pipeline processes.

        Arguments:
        None

        Return:
        None
        """
        self.queue_lock.lock()
        try:
            if self.queue.empty():
                self.queue_com['status'].put([
                    'Waiting {0} {1}'.format(
                        self.queue.qsize(),
                        self.shared_dict_typ['file_number']
                        ),
                    self.name,
                    'orange'
                    ])
                QThread.sleep(5)
                return None
            else:
                pass
        except Exception:
            return None
        finally:
            self.queue_lock.unlock()

        # Get new file
        self.queue_lock.lock()
        try:
            self.queue_com['status'].put([
                'Running {0} {1}'.format(
                    self.queue.qsize(),
                    self.shared_dict_typ['file_number']
                    ),
                self.name,
                'green'
                ])
            root_name = self.remove_from_queue()
            QThread.sleep(1)
        except Exception:
            return None
        finally:
            self.queue_lock.unlock()

        # Set for every process a method and the right lost_connection name
        method_dict = {
            'Copy': {
                'method': self.run_copy,
                'lost_connect': 'full_transphire'
                },
            'Motion': {
                'method': self.run_motion,
                'lost_connect': 'full_transphire'
                },
            'CTF': {
                'method': self.run_ctf,
                'lost_connect': 'full_transphire'
                },
            'Compress': {
                'method': self.run_compress,
                'lost_connect': 'full_transphire'
                },
            'Copy_work': {
                'method': self.run_copy_extern,
                'lost_connect': 'full_work'
                },
            'Copy_backup': {
                'method': self.run_copy_extern,
                'lost_connect': 'full_backup'
                },
            'Copy_hdd': {
                'method': self.run_copy_extern,
                'lost_connect': 'full_hdd'
                }
            }

        try:
            method_dict[self.typ]['method'](
                root_name=root_name,
                )
        except FileNotFoundError as err:
            self.add_to_queue(aim=self.typ, root_name=root_name)
            self.write_error(msg=tb.format_exc(), root_name=root_name)
            if 'Check Startnumber' in str(err):
                pass
            else:
                self.lost_connection(
                    typ='lost_input_meta'
                    )
                self.lost_connection(
                    typ='lost_input_frames'
                    )
        except BlockingIOError:
            self.add_to_queue(aim=self.typ, root_name=root_name)
            print('!!! BlockingIOError !!! \n')
            msg = tb.format_exc()
            self.write_error(msg=msg, root_name=root_name)
            if 'termios.error' not in msg:
                message = ''.join([
                    'BlockingIOError occured in {0}!'.format(self.name),
                    'The process will continue, but check the error file!',
                    '\n{0}'.format(self.shared_dict_typ['error_file'])
                    ])
                self.queue_com['notification'].put(message)
                if ti.time() - self.time_last_error > 1800:
                    self.queue_com['error'].put(message)
                    self.time_last_error = ti.time()
                else:
                    pass
                self.shared_dict_typ['unknown_error'] = True
            else:
                pass
        except IOError:
            self.add_to_queue(aim=self.typ, root_name=root_name)
            self.write_error(msg=tb.format_exc(), root_name=root_name)
            self.lost_connection(
                typ=method_dict[self.typ]['lost_connect']
                )
        except Exception:
            self.add_to_queue(aim=self.typ, root_name=root_name)
            print('!!! UNKNOWN !!! \n')
            msg = tb.format_exc()
            self.write_error(msg=msg, root_name=root_name)
            if 'termios.error' not in msg:
                message = ''.join([
                    'Unknown error occured in {0}!'.format(self.name),
                    'The process will continue, but check the error file!',
                    '\n{0}'.format(self.shared_dict_typ['error_file'])
                    ])
                self.queue_com['notification'].put(message)
                if ti.time() - self.time_last_error > 1800:
                    self.queue_com['error'].put(message)
                    self.time_last_error = ti.time()
                else:
                    pass
                self.shared_dict_typ['unknown_error'] = True
            else:
                pass
        else:
            self.remove_from_queue_file(root_name)
            self.queue_lock.lock()
            try:
                self.add_to_queue_file(
                    root_name=root_name,
                    file_name=self.shared_dict['typ'][self.typ]['done_file'],
                    )
            except Exception:
                raise
            finally:
                self.queue_lock.unlock()
            if self.typ == 'Copy':
                pass
            else:
                self.queue_lock.lock()
                self.shared_dict_typ['file_number'] += 1
                self.queue_lock.unlock()

    def write_error(self, msg, root_name):
        """
        Write to error file.

        Arguments:
        msg - Message to send.
        root_name - File that was processed while the error occured.

        Return:
        None
        """
        self.shared_dict_typ['error_lock'].lock()
        try:
            local = ti.localtime()
            print('\n{0}/{1}/{2}-{3}:{4}:{5}\t'.format(*local[0:6]))
            print(root_name)
            print('New error message in error file: {0}'.format(
                self.shared_dict_typ['error_file']
                ))
            with open(self.shared_dict_typ['error_file'], 'a') as append:
                append.write('{0}/{1}/{2}-{3}:{4}:{5}\t'.format(*local[0:6]))
                append.write('{0}\n'.format(self.typ))
                append.write('{0}\n'.format(root_name))
                if self.password:
                    append.write('{0}\n\n\n'.format(
                        str(msg).replace(self.password, 'SUDOPASSWORD')
                        ))
                else:
                    append.write('{0}\n\n\n'.format(str(msg)))
        except Exception:
            raise
        finally:
            self.shared_dict_typ['error_lock'].unlock()

    def remove_from_queue(self):
        """
        Remove item from queue.

        Arguments:
        None

        Return:
        Name removed from the queue.
        """
        return self.queue.get(block=False)

    @staticmethod
    def add_to_queue_file(root_name, file_name):
        """
        Add item to queue_file.

        Arguments:
        root_name - Name to add
        file_name - File to append to

        Return:
        Name removed from the queue.
        """
        is_present = False
        try:
            with open(file_name, 'r') as read:
                is_present = bool(root_name in read.read())
        except FileNotFoundError:
            pass

        if not is_present:
            with open(file_name, 'a') as append:
                append.write('{0}\n'.format(root_name))
        else:
            pass

    def add_to_queue(self, aim, root_name):
        """
        Add item to queue.

        Arguments:
        aim - Aim queue
        root_name - Name to add

        Return:
        None
        """
        self.shared_dict['typ'][aim]['queue_lock'].lock()
        try:
            self.shared_dict['queue'][aim].put(root_name, block=False)
            self.add_to_queue_file(
                root_name=root_name,
                file_name=self.shared_dict['typ'][aim]['save_file']
                )
        except Exception:
            raise
        finally:
            self.shared_dict['typ'][aim]['queue_lock'].unlock()

    def lost_connection(self, typ):
        """
        Handle lost connection errors.

        Arguments:
        typ - Typ of the error

        Return:
        None
        """
        if typ == 'Motion':
            message = ''.join([
                '{0}: An error occured! '.format(self.name),
                'MotionCor2 crashed or em-transfer is full ',
                'or SumMove/MotionCor2 path is not specified correctly! ',
                'MorionCor2 processes stopped! ',
                'It is possible that the PC needs to be rebooted! ',
                'Please check the error file!'
                ])
        else:
            message = ''.join([
                '{0}: An error occured!'.format(self.name),
                'Lost connection or target device out of space!',
                'Please check the error file!'
                ])
        self.shared_dict_typ[typ] = True
        message = '{0} ({1}, {2})'.format(
            message,
            self.name,
            self.shared_dict_typ['error_file']
            )
        self.queue_com['notification'].put(message)
        if ti.time() - self.time_last_error > 1800:
            self.queue_com['error'].put(message)
            self.time_last_error = ti.time()
        else:
            pass

    def remove_from_queue_file(self, root_name):
        """
        Remove the files from the queue file.

        Arguments:
        root_name - Name of the file to delete

        Return:
        None
        """
        self.queue_lock.lock()
        try:
            useable_lines = []
            try:
                with open(self.shared_dict_typ['save_file'], 'r') as read:
                    lines = [line.rstrip() for line in read.readlines()]
            except FileNotFoundError:
                useable_lines = []
            else:
                for line in lines:
                    if root_name != line:
                        useable_lines.append(line)
                    else:
                        pass

            with open(self.shared_dict_typ['save_file'], 'w') as write:
                write.write('{0}\n'.format('\n'.join(useable_lines)))
        except Exception:
            raise
        finally:
            self.queue_lock.unlock()

    def run_software_meta(self, directory):
        """
        Copy meta files produces by the collection software.

        Arguments:
        directory - Start directory for recursive search.

        Return:
        None
        """
        file_list = []
        file_list = self.recursive_search(
            directory=directory,
            file_list=file_list,
            find_meta=True
            )

        for entry in file_list:
            if self.stop:
                break
            else:
                pass
            file_path = entry[len(directory)+1:]
            file_name = os.path.join(self.settings['software_meta_folder'], file_path)
            tu.mkdir_p(os.path.dirname(file_name))
            tu.copy(entry, file_name)
            # Add to queue

        folder_name = os.path.basename(self.settings['software_meta_folder'])
        tar_file = '{0}.tar'.format(self.settings['software_meta_folder'])
        with tarfile.open(tar_file, 'w') as tar:
            tar.add(self.settings['software_meta_folder'], arcname=folder_name)

        for aim in self.content_settings['aim']:
            *compare, aim_name = aim.split(':')
            var = True
            for typ in compare:
                name = typ.split('!')[-1]
                if typ.startswith('!'):
                    if self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
                else:
                    if not self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
            if var:
                self.add_to_queue(aim=aim_name, root_name=tar_file)
            else:
                pass


    def run_find(self):
        """
        Find files

        Arguments:
        None

        Return:
        None
        """
        self.queue_lock.lock()
        file_list = []
        try:
            file_list = self.recursive_search(
                directory=self.settings['General']['Search path meta'],
                file_list=file_list,
                find_meta=False
                )
        except Exception:
            raise
        finally:
            self.queue_lock.unlock()

        data = np.empty(
            len(file_list),
            dtype=[('root', '|S200'), ('date', '<i8'), ('time', '<i8')]
            )

        for idx, root_name in enumerate(file_list):
            hole, grid_number, spot1, spot2, date, time = \
                tus.extract_time_and_grid_information(
                    root_name=root_name,
                    settings=self.settings,
                    queue_com=self.queue_com,
                    name=self.name
                    )
            del spot1, spot2, hole, grid_number
            data[idx]['root'] = root_name
            data[idx]['date'] = int(date)
            data[idx]['time'] = int(time)

        data = np.sort(data, order=['date', 'time'])
        for root_name in data['root']:
            for aim in self.content_settings['aim']:
                *compare, aim_name = aim.split(':')
                var = True
                for typ in compare:
                    name = typ.split('!')[-1]
                    if typ.startswith('!'):
                        if self.settings['Copy'][name] == 'False':
                            continue
                        else:
                            var = False
                            break
                    else:
                        if not self.settings['Copy'][name] == 'False':
                            continue
                        else:
                            var = False
                            break
                if var:
                    self.add_to_queue(
                        aim=aim_name,
                        root_name=root_name.decode('utf-8')
                        )
                else:
                    pass

    def recursive_search(self, directory, file_list, find_meta):
        """
        Find files in a recursive search.

        directory - Directory to search files.
        file_list - List of files that have been found.
        find_meta - Find meta data flag.

        Returns:
        List of files found
        """
        if not os.path.exists(directory):
            raise FileNotFoundError('Find directory does not exist')
        else:
            pass

        files_in_dir = sorted(glob.glob('{0}/*'.format(directory)))

        for entry_dir in files_in_dir:
            if self.stop:
                break
            elif os.path.isdir(entry_dir):
                file_list = self.recursive_search(
                    directory=entry_dir,
                    file_list=file_list,
                    find_meta=find_meta
                    )
            elif find_meta:
                if os.path.isfile(entry_dir) and 'Data' not in entry_dir:
                    file_list.append(entry_dir)
                else:
                    continue
            elif os.path.isfile(entry_dir) and \
                    'Data' in entry_dir and \
                    entry_dir.endswith('.jpg'):
                root_name = entry_dir[:-len('.jpg')]
                self.shared_dict_typ['bad_lock'].lock()
                try:
                    if root_name in self.shared_dict['bad'][self.typ]:
                        continue
                    else:
                        pass
                except Exception:
                    raise
                finally:
                    self.shared_dict_typ['bad_lock'].unlock()

                frames_root = root_name.replace(
                    self.settings['General']['Search path meta'],
                    self.settings['General']['Search path frames'],
                    )
                compare_name = frames_root[:-len('_19911213_2019')]

                frames = tus.find_frames(
                    frames_root=frames_root,
                    compare_name=compare_name,
                    settings=self.settings,
                    queue_com=self.queue_com,
                    name=self.name,
                    write_error=self.write_error
                    )
                if frames is None:
                    self.shared_dict_typ['bad_lock'].lock()
                    try:
                        if root_name not in self.shared_dict['bad'][self.typ]:
                            self.shared_dict['bad'][self.typ].append(root_name)
                        else:
                            pass
                    except Exception:
                        raise
                    finally:
                        self.shared_dict_typ['bad_lock'].unlock()
                    continue
                elif not frames:
                    continue
                else:
                    pass

                self.shared_dict['typ'][self.content_settings['group']]['share_lock'].lock()
                try:
                    if root_name in self.shared_dict['share'][self.content_settings['group']]:
                        continue
                    else:
                        self.time_last = ti.time()
                        self.notification_send = False
                        file_list.append(root_name)
                        self.shared_dict['share'][self.content_settings['group']].append(
                            root_name
                            )
                except Exception:
                    raise
                finally:
                    self.shared_dict['typ'][self.content_settings['group']]['share_lock'].unlock()
            else:
                continue

        return file_list

    def run_copy(self, root_name):
        """
        Copy found files to em-transfer.

        root_name - Root name of file to copy.

        Returns:
        None
        """
        frames_root = root_name.replace(
            self.settings['General']['Search path meta'],
            self.settings['General']['Search path frames'],
            )
        frames, compare_name_meta = tus.find_related_frames_to_jpg(
            frames_root=frames_root,
            root_name=root_name,
            settings=self.settings,
            queue_com=self.queue_com,
            name=self.name
            )

        overall_file_size = 0
        for frame in frames:
            overall_file_size += os.path.getsize(frame)

        self.shared_dict_typ['count_lock'].lock()
        try:
            self.shared_dict_typ['file_number'] += 1
            with open(self.shared_dict_typ['number_file'], 'w') as write:
                write.write(str(self.shared_dict_typ['file_number']))
            if self.settings['General']['Rename micrographs'] == 'True':
                new_name_stack = '{0}/{1}{2:0{3}d}{4}'.format(
                    self.settings['stack_folder'],
                    self.settings['General']['Rename prefix'],
                    self.shared_dict_typ['file_number'],
                    len(self.settings['General']['Estimated mic number']),
                    self.settings['General']['Rename suffix']
                    )
                new_name_meta = '{0}/{1}{2:0{3}d}{4}'.format(
                    self.settings['meta_folder'],
                    self.settings['General']['Rename prefix'],
                    self.shared_dict_typ['file_number'],
                    len(self.settings['General']['Estimated mic number']),
                    self.settings['General']['Rename suffix']
                    )
            else:
                new_name_stack = os.path.join(
                    self.settings['stack_folder'],
                    root_name.split('/')[-1]
                    )
                new_name_meta = os.path.join(
                    self.settings['meta_folder'],
                    root_name.split('/')[-1]
                    )
        except Exception:
            raise
        finally:
            self.shared_dict_typ['count_lock'].unlock()

        if frames:
            check_name = root_name.split('/')[-1]
            if self.already_in_translation_file(root_name=check_name):
                message = \
                    '{0}: In queue, but already copied! Skip!'.format(self.name)
                self.write_error(msg=message, root_name=root_name)
                return None
            else:
                pass
        else:
            self.stop = True
            message = '{0}: No frames found!'.format(self.name)
            self.queue_com['error'].put(message, self.name)
            self.queue_com['notification'].put(message)
            self.write_error(msg=message, root_name=root_name)
            raise IOError(message)

        if os.path.exists('{0}.jpg'.format(new_name_meta)):
            self.stop = True
            if os.path.exists(self.shared_dict_typ['done_file']):
                self.queue_lock.lock()
                try:
                    with open(self.shared_dict_typ['done_file'], 'r') as read:
                        self.shared_dict_typ['file_number'] = len(read.readlines())
                except Exception:
                    raise
                finally:
                    self.queue_lock.unlock()
            else:
                self.shared_dict_typ['file_number'] = int(
                    self.settings['General']['Start number']
                    )
            message = '{0}: File {1} already exists!\n'.format(
                self.name,
                new_name_meta
                ) + \
              'Check Startnumber!'
            self.queue_com['notification'].put(message)
            raise FileNotFoundError(message)
        else:
            pass

        if overall_file_size > \
                sh.disk_usage(self.settings['project_folder']).free:
            self.stop = True
            message = '{0}: Not enough space in project folder'.format(
                self.name
                )
            self.queue_com['notification'].put(message)
            raise IOError(message)
        else:
            pass

        new_stack = '{0}.{1}'.format(
            new_name_stack,
            self.settings['General']['Output extension']
            )

        command = tus.get_copy_command_for_frames(
            settings=self.settings,
            queue_com=self.queue_com,
            name=self.name
            )

        newstack_command = '{0} {1} {2}'.format(
            command,
            ' '.join(frames),
            new_stack
            )

        # Create the stack
        file_stdout = '{0}.log'.format(new_name_stack)
        file_stderr = '{0}.err'.format(new_name_stack)
        with open(file_stdout, 'w') as out:
            out.write(newstack_command)
            with open(file_stderr, 'w') as err:
                start_time = ti.time()
                sp.Popen(newstack_command.split(), stdout=out, stderr=err).wait()
                stop_time = ti.time()
                out.write('\nTime: {0} sec'.format(stop_time - start_time))

        tus.check_outputs(
            zero_list=[file_stderr],
            non_zero_list=[file_stdout, new_stack],
            folder=self.settings['stack_folder'],
            command=newstack_command
            )

        all_files = tus.find_all_files(
            root_name=root_name,
            compare_name_meta=compare_name_meta,
            settings=self.settings,
            queue_com=self.queue_com,
            name=self.name
            )

        xml_file = None
        log_files = []
        for file_entry in all_files:
            extension = file_entry.split('.')[-1]
            if file_entry in frames:
                continue
            elif extension == 'mrc':
                name = '{0}_krios_sum'.format(new_name_meta)
            elif extension == 'dm4' and 'gain' in file_entry:
                name = '{0}_gain'.format(new_name_meta)
            else:
                name = new_name_meta

            new_file = '{0}.{1}'.format(name, extension)

            if extension == 'xml':
                xml_file = new_file
            else:
                pass

            tu.copy('{0}'.format(file_entry), new_file)
            log_files.append(new_file)

        tus.check_outputs(
            zero_list=[],
            non_zero_list=log_files,
            folder=self.settings['meta_folder'],
            command='copy'
            )

        log_files.extend([file_stdout, file_stderr])
        for file_entry in all_files:
            try:
                os.remove(file_entry)
            except IOError:
                self.write_error(msg=tb.format_exc(), root_name=file_entry)
                raise

        self.append_to_translate(
            root_name=root_name,
            new_name=new_name_meta,
            xml_file=xml_file
            )

        self.shared_dict['typ'][self.content_settings['group']]['share_lock'].lock()
        try:
            self.shared_dict['share'][self.content_settings['group']].remove(root_name)
        except Exception:
            raise
        finally:
            self.shared_dict['typ'][self.content_settings['group']]['share_lock'].unlock()

        for aim in self.content_settings['aim']:
            *compare, aim_name = aim.split(':')
            var = True
            for typ in compare:
                name = typ.split('!')[-1]
                if typ.startswith('!'):
                    if self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
                else:
                    if not self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
            if var:
                if '!Compress data' in compare or \
                        'Compress data' in compare or \
                        'Motion' in compare or \
                        'CTF_frames' in compare:
                    self.add_to_queue(aim=aim_name, root_name=new_stack)
                else:
                    for log_file in log_files:
                        self.add_to_queue(aim=aim_name, root_name=log_file)
            else:
                pass

    def already_in_translation_file(self, root_name):
        """
        Check, if the root_name already exists in the translation file.

        root_name - root_name

        Returns:
        True, if root_name in translation file.
        """
        self.shared_dict['translate_lock'].lock()
        try:
            content_translation_file = np.genfromtxt(
                os.path.join(
                    self.settings['project_folder'],
                    'Translation_file.txt'
                    ),
                usecols=0,
                dtype=str
                )
        except OSError:
            return False
        except Exception:
            raise
        else:
            return bool(root_name in content_translation_file)
        finally:
            self.shared_dict['translate_lock'].unlock()

    def append_to_translate(self, root_name, new_name, xml_file):
        """
        Write to the translation file.

        root_name - Root name of the file
        new_name - New name of the file

        Returns:
        None
        """
        file_name = os.path.join(
            self.settings['project_folder'],
            'Translation_file.txt'
            )

        if os.path.exists(file_name):
            first_entry = []
        else:
            first_entry = [
                '_pipeRootName',
                '_pipeNewName',
                '_pipeHoleNumber',
                '_pipeSpotNumber',
                '_pipeDate',
                '_pipeTime',
                '_pipeGridNumber',
                ]

        self.shared_dict['translate_lock'].lock()
        try:
            try:
                hole, grid_number, spot1, spot2, date, time = \
                    tus.extract_time_and_grid_information(
                        root_name=root_name,
                        settings=self.settings,
                        queue_com=self.queue_com,
                        name=self.name
                        )
                new_file = os.path.basename(new_name)

                key = '{0}{1}'.format(spot1, spot2)
                if key not in self.shared_dict_typ['spot_dict']:
                    self.shared_dict_typ['spot'] += 1
                    self.shared_dict_typ['spot_dict'][key] = \
                        self.shared_dict_typ['spot']
                    with open(
                            '{0}/.spot_dict'.format(
                                self.settings['project_folder']
                                ),
                            'a'
                            ) as append:
                        append.write('{0}\t{1}\n'.format(
                            key,
                            self.shared_dict_typ['spot']
                            ))

                entries = [
                    os.path.basename(root_name),
                    new_file,
                    hole,
                    self.shared_dict_typ['spot_dict'][key],
                    date,
                    time,
                    grid_number
                    ]

                if xml_file is None:
                    pass
                else:
                    try:
                        with open(xml_file, 'r') as read:
                            lines = read.read()
                    except IOError:
                        pass
                    else:
                        try:
                            value_x = re.match('.*<X>(.*)</X>.*', lines).group(1)
                            value_y = re.match('.*<Y>(.*)</Y>.*', lines).group(1)
                            value_z = re.match('.*<Z>(.*)</Z>.*', lines).group(1)
                            defocus = re.match('.*<Defocus>(.*)</Defocus>.*', lines).group(1)
                        except AttributeError:
                            pass
                        else:
                            entries.append(value_x)
                            entries.append(value_y)
                            entries.append(value_z)
                            entries.append(defocus)
                            if first_entry:
                                first_entry.append('_pipeCoordX')
                                first_entry.append('_pipeCoordY')
                                first_entry.append('_pipeCoordZ')
                                first_entry.append('_pipeDefocusMicroscope')
                            else:
                                pass

            except ValueError:
                if first_entry:
                    first_entry = [
                        '_pipeRootName',
                        '_pipeRootDir',
                        '_pipeNewName'
                        ]
                else:
                    pass
                entries = [
                    os.path.basename(root_name),
                    os.path.dirname(root_name),
                    new_name
                    ]

            template = '{0}\n'
            with open(file_name, 'a') as write:
                if first_entry:
                    write.write(template.format('\n'.join(first_entry)))
                else:
                    pass
                write.write(
                    template.format(
                        '\t'.join(
                            ['{0}'.format(entry) for entry in entries]
                            )
                        )
                    )
        except Exception:
            raise
        finally:
            self.shared_dict['translate_lock'].unlock()

        if not self.settings['Copy']['Copy to work'] == 'False':
            self.add_to_queue(aim='Copy_work', root_name=file_name)
        else:
            pass
        if not self.settings['Copy']['Copy to HDD'] == 'False':
            self.add_to_queue(aim='Copy_hdd', root_name=file_name)
        else:
            pass
        if not self.settings['Copy']['Copy to backup'] == 'False':
            self.add_to_queue(aim='Copy_backup', root_name=file_name)
        else:
            pass

    def run_motion(self, root_name):
        """
        Do the motion correction.

        root_name - Root name of the micrograph.

        Returns:
        None
        """
        file_input = root_name
        root_name, _ = os.path.splitext(file_input)
        file_dw_post_move = None
        file_stack = None
        queue_dict = {}
        for motion_idx, key in enumerate(self.settings['motion_frames']):
            # The current settings that we work with
            motion_frames = copy.deepcopy(self.settings['motion_frames'][key])
            queue_dict[motion_idx] = {'log': [], 'sum': [], 'sum_dw': []}

            # Abort if frames out of range
            if motion_frames['first'] > \
                    int(self.settings['General']['Number of frames']) or \
                    motion_frames['last'] > \
                    int(self.settings['General']['Number of frames']):
                print('First:{0} Last:{1} not valid! Skip!\n'.format(
                    motion_frames['first'],
                    motion_frames['last']
                    ))
                continue
            else:
                pass

            # Create an unblur shift file
            file_shift = os.path.join(
                self.settings['motion_folder'],
                '.{0}_shift'.format(self.name)
                )

            # Rename the last frame
            if motion_frames['last'] == -1:
                do_dw = tum.get_motion_default(
                    settings=self.settings,
                    motion_frames=motion_frames,
                    queue_com=self.queue_com,
                    name=self.name
                    )
                zeros_list = [
                    '0.0' for i in range(
                        motion_frames['last'] - motion_frames['first'] - 1
                        )
                    ]
                with open(file_shift, 'w') as write:
                    write.write('{0}\n{0}\n'.format('\t'.join(zeros_list)))
            else:
                do_dw = False

            # Folder name for the current settings
            if do_dw:
                suffix = 'with_DW'
            else:
                suffix = 'without_DW'
            output_folder_name = '{0}_{1}_{2}'.format(
                motion_frames['first'],
                motion_frames['last'],
                suffix
                )
            output_logfile = '{0}_log'.format(output_folder_name)

            # Create the folders
            # Variables
            output_transfer_root = os.path.join(
                self.settings['motion_folder'],
                output_folder_name
                )
            output_transfer_log = os.path.join(
                self.settings['motion_folder'],
                output_logfile
                )
            # Scratch
            output_transfer_scratch_root = os.path.join(
                self.settings['scratch_motion_folder'],
                output_folder_name
                )
            output_transfer_log_scratch = os.path.join(
                self.settings['scratch_motion_folder'],
                output_logfile
                )


            output_dw = os.path.join(output_transfer_root, 'DW')
            output_transfer = os.path.join(output_transfer_root, 'Non_DW')
            output_transfer_scratch = os.path.join(
                output_transfer_scratch_root,
                'Non_DW'
                )

            # Create folders if they do not exist
            self.queue_lock.lock()
            try:
                tu.mkdir_p(output_transfer_root)
                tu.mkdir_p(output_transfer)
                tu.mkdir_p(output_transfer_log)
                tu.mkdir_p(output_transfer_scratch)
                tu.mkdir_p(output_transfer_log_scratch)
            except Exception:
                raise
            finally:
                self.queue_lock.unlock()

            # Remove the path from the name
            file_name = os.path.basename(root_name)

            # Scratch output
            file_output_scratch = os.path.join(
                output_transfer_scratch,
                '{0}.mrc'.format(file_name)
                )
            file_log_scratch = os.path.join(
                output_transfer_log_scratch,
                '{0}.mrc'.format(file_name)
                )
            file_stdout_scratch = '{0}-output.log'.format(
                file_log_scratch
                )
            file_stderr_scratch = '{0}-error.log'.format(
                file_log_scratch
                )

            non_zero_list_scratch = [
                file_output_scratch,
                file_stdout_scratch
                ]
            zero_list_scratch = [file_stderr_scratch]

            # Project output
            file_output = os.path.join(
                output_transfer,
                '{0}.mrc'.format(file_name)
                )
            file_log = os.path.join(
                output_transfer_log,
                '{0}.mrc'.format(file_name)
                )
            file_stdout = '{0}-output.log'.format(
                file_log
                )
            file_stderr = '{0}-error.log'.format(
                file_log
                )
            file_frc = '{0}-frc.log'.format(
                file_log
                )

            non_zero_list = [
                file_output,
                file_stdout
                ]
            zero_list = [file_stderr]

            # Create the commands
            if motion_idx == 0:
                # DW folder
                self.queue_lock.lock()
                try:
                    tu.mkdir_p(output_dw)
                except Exception:
                    raise
                finally:
                    self.queue_lock.unlock()

                # Files
                file_stack = os.path.join(
                    output_transfer_scratch,
                    '{0}_Stk.mrc'.format(file_name)
                    )
                file_dw_post_move = os.path.join(
                    output_dw,
                    '{0}.mrc'.format(file_name)
                    )
                file_dw_pre_move = tum.get_dw_file_name(
                    output_transfer_scratch=output_transfer_scratch,
                    file_name=file_name,
                    settings=self.settings,
                    queue_com=self.queue_com,
                    name=self.name
                    )
                command = tum.get_motion_command(
                    file_input=file_input,
                    file_output_scratch=file_output_scratch,
                    file_log_scratch=file_log_scratch,
                    queue_com=self.queue_com,
                    name=self.name,
                    settings=self.settings,
                    )

                with open(file_stdout_scratch, 'w') as out:
                    out.write(command)
                    with open(file_stderr_scratch, 'w') as err:
                        start_time = ti.time()
                        sp.Popen(command.split(), stdout=out, stderr=err).wait()
                        stop_time = ti.time()
                        out.write('\nTime: {0} sec'.format(stop_time - start_time))

                # Move DW file
                if do_dw:
                    non_zero_list_scratch.append(file_dw_pre_move)
                    non_zero_list.append(file_dw_post_move)
                else:
                    pass

            else:
                command = tum.create_sum_movie_command(
                    motion_frames=motion_frames,
                    file_input=file_stack,
                    file_output=file_output_scratch,
                    file_shift=file_shift,
                    file_frc=file_frc,
                    settings=self.settings,
                    queue_com=self.queue_com,
                    name=self.name
                    )
                with open(file_stdout_scratch, 'w') as out:
                    out.write(command)
                    with open(file_stderr_scratch, 'w') as err:
                        start_time = ti.time()
                        sp.Popen(command, shell=True, stdout=out, stderr=err).wait()
                        stop_time = ti.time()
                        out.write('\nTime: {0} sec'.format(stop_time - start_time))

                non_zero_list.append(file_frc)
                self.queue_lock.lock()
                try:
                    for entry_temp in glob.glob('.SumMovie*'):
                        os.remove(entry_temp)
                except Exception:
                    raise
                finally:
                    self.queue_lock.unlock()

            # Sanity check
            log_files_scratch = glob.glob('{0}0*'.format(file_log_scratch))
            non_zero_list_scratch.extend(log_files_scratch)
            tus.check_outputs(
                zero_list=zero_list_scratch,
                non_zero_list=non_zero_list_scratch,
                folder=output_transfer_log_scratch,
                command=command
                )

            if do_dw:
                tu.copy(file_dw_pre_move, file_dw_post_move)
            else:
                pass

            if os.path.realpath(self.settings['scratch_motion_folder']) != \
                    os.path.realpath(self.settings['motion_folder']):
                self.queue_lock.lock()
                try:
                    tu.copy(file_output_scratch, file_output)
                    tu.copy(file_stdout_scratch, file_stdout)
                    tu.copy(file_stderr_scratch, file_stderr)
                    for file_name_log in log_files_scratch:
                        name = os.path.basename(file_name_log)
                        new_name = os.path.join(output_transfer_log, name)
                        tu.copy(file_name_log, new_name)
                except Exception:
                    raise
                finally:
                    self.queue_lock.unlock()
            else:
                pass

            log_files = glob.glob('{0}0*'.format(file_log))
            non_zero_list.extend(log_files)
            tus.check_outputs(
                zero_list=zero_list,
                non_zero_list=non_zero_list,
                folder=self.settings['motion_folder'],
                command='copy'
                )

            copied_files = zero_list_scratch + non_zero_list_scratch
            for file_entry in copied_files:
                try:
                    os.remove(file_entry)
                except IOError:
                    self.write_error(msg=tb.format_exc(), root_name=file_entry)
                    raise

            queue_dict[motion_idx]['sum'].append(file_output)
            for file_name_log in glob.glob('{0}*'.format(file_log)):
                queue_dict[motion_idx]['log'].append(file_name_log)
            if do_dw:
                queue_dict[motion_idx]['sum_dw'].append(file_dw_post_move)
            else:
                pass

        for motion_idx in queue_dict:
            for aim in self.content_settings['aim']:
                *compare, aim_name = aim.split(':')
                var = True
                for typ in compare:
                    name = typ.split('!')[-1]
                    if typ.startswith('!'):
                        if self.settings['Copy'][name] == 'False':
                            continue
                        else:
                            var = False
                            break
                    else:
                        if not self.settings['Copy'][name] == 'False':
                            continue
                        else:
                            var = False
                            break
                if var:
                    sum_files = queue_dict[motion_idx]['sum']
                    log_files = queue_dict[motion_idx]['log']
                    sum_dw_files = queue_dict[motion_idx]['sum_dw']
                    if '!Compress data' in compare:
                        if motion_idx == 0:
                            self.add_to_queue(aim=aim_name, root_name=file_input)
                        else:
                            pass
                    elif 'Compress data' in compare:
                        if motion_idx == 0:
                            self.add_to_queue(aim=aim_name, root_name=file_input)
                        else:
                            pass
                    elif 'CTF_frames' in compare:
                        if motion_idx == 0:
                            for file_name in sum_files:
                                self.add_to_queue(
                                    aim=aim_name,
                                    root_name='{0};;;{1}'.format(file_name, file_input)
                                    )
                        else:
                            pass
                    elif 'CTF_sum' in compare:
                        if motion_idx == 0:
                            for file_name in sum_files:
                                self.add_to_queue(aim=aim_name, root_name=file_name)
                        else:
                            pass
                    else:
                        for file_name in sum_files:
                            self.add_to_queue(aim=aim_name, root_name=file_name)
                        for file_name in log_files:
                            self.add_to_queue(aim=aim_name, root_name=file_name)
                        for file_name in sum_dw_files:
                            self.add_to_queue(aim=aim_name, root_name=file_name)
                else:
                    pass

        # Plot Motion information
        self.queue_lock.lock()
        self.queue_com['plot_motion'].put(True)
        try:
            os.remove(file_stack)
        except Exception:
            raise
        finally:
            self.queue_lock.unlock()

    def run_ctf(self, root_name):
        """
        Run CTF estimation.

        root_name - name of the file to process.

        Returns:
        None
        """
        if ';;;' in root_name:
            sum_file, file_input = root_name.split(';;;')
        else:
            sum_file = root_name
            file_input = root_name
        root_name, _ = os.path.splitext(file_input)

        # New name
        file_name = os.path.basename(root_name)
        new_name = os.path.join(
            self.settings['ctf_folder'],
            '{0}.mrc'.format(file_name)
            )

        # Create the command
        command, check_files = tuc.get_ctf_command(
            file_input=file_input,
            new_name=new_name,
            settings=self.settings,
            queue_com=self.queue_com,
            name=self.name
            )

        # Log files
        out_file = os.path.join(
            self.settings['ctf_folder'],
            '{0}.log'.format(file_name)
            )
        err_file = os.path.join(
            self.settings['ctf_folder'],
            '{0}.err'.format(file_name)
            )

        # Run the command
        with open(out_file, 'w') as out:
            out.write(command)
            with open(err_file, 'w') as err:
                start_time = ti.time()
                sp.Popen(command, shell=True, stdout=out, stderr=err).wait()
                stop_time = ti.time()
                out.write('\nTime: {0} sec'.format(stop_time - start_time))

        zero_list = [err_file]
        non_zero_list = [out_file]
        non_zero_list.extend(check_files)

        root_path = os.path.join(os.path.dirname(root_name), file_name)
        log_files, copied_log_files = tuc.find_logfiles(
            root_path=root_path,
            file_name=file_name,
            settings=self.settings,
            queue_com=self.queue_com,
            name=self.name
            )

        try:
            log_files.remove(err_file)
            copied_log_files.remove(err_file)
        except ValueError:
            pass

        tus.check_outputs(
            zero_list=zero_list,
            non_zero_list=non_zero_list+log_files,
            folder=self.settings['ctf_folder'],
            command=command
            )

        for old_file, new_file in zip(log_files, copied_log_files):
            if os.path.realpath(old_file) != os.path.realpath(new_file):
                tu.copy(old_file, new_file)
            else:
                pass

        tus.check_outputs(
            zero_list=[],
            non_zero_list=copied_log_files,
            folder=self.settings['ctf_folder'],
            command=command
            )

        for old_file, new_file in zip(log_files, copied_log_files):
            if os.path.realpath(old_file) != os.path.realpath(new_file):
                os.remove(old_file)
            else:
                pass

        copied_log_files.extend(non_zero_list)
        copied_log_files.extend(zero_list)
        copied_log_files = list(set(copied_log_files))

        # Combine output files
        output_name_partres, output_name_star = tuc.combine_ctf_outputs(
            root_path=root_path,
            file_name=file_name,
            settings=self.settings,
            queue_com=self.queue_com,
            shared_dict=self.shared_dict,
            name=self.name,
            sum_file=sum_file
            )

        if not self.settings['Copy']['Copy to work'] == 'False':
            self.add_to_queue(aim='Copy_work', root_name=output_name_partres)
            self.add_to_queue(aim='Copy_work', root_name=output_name_star)
        else:
            pass
        if not self.settings['Copy']['Copy to HDD'] == 'False':
            self.add_to_queue(aim='Copy_hdd', root_name=output_name_partres)
            self.add_to_queue(aim='Copy_hdd', root_name=output_name_star)
        else:
            pass
        if not self.settings['Copy']['Copy to backup'] == 'False':
            self.add_to_queue(aim='Copy_backup', root_name=output_name_partres)
            self.add_to_queue(aim='Copy_backup', root_name=output_name_star)
        else:
            pass

        # Add to queue
        for aim in self.content_settings['aim']:
            *compare, aim_name = aim.split(':')
            var = True
            for typ in compare:
                name = typ.split('!')[-1]
                if typ.startswith('!'):
                    if self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
                else:
                    if not self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
            if var:
                if '!Compress data' in compare or 'Compress data' in compare:
                    self.add_to_queue(aim=aim_name, root_name=file_input)
                else:
                    for log_file in copied_log_files:
                        self.add_to_queue(aim=aim_name, root_name=log_file)
            else:
                pass

        # Plot CTF information
        self.queue_com['plot_ctf'].put(True)

    def run_compress(self, root_name):
        """
        Compress stack.

        root_name - Name of the file to compress

        Returns:
        None
        """
        if self.settings['compress_folder'] in root_name:
            name, _ = os.path.splitext(root_name)
            new_name = root_name
            log_file = '{0}.log'.format(name)
            err_file = '{0}.err'.format(name)
            tus.check_outputs(
                zero_list=[err_file],
                non_zero_list=[log_file, new_name],
                folder=self.settings['compress_folder'],
                command='just check'
                )

        else:
            # New name
            new_root_name, extension = os.path.splitext(os.path.basename(root_name))

            new_name = os.path.join(
                    self.settings['compress_folder'],
                    '{0}.tiff'.format(new_root_name)
                    )

            # Create the command
            if extension == '.mrc':
                command = '{0} -s -c lzw {1} {2}'.format(
                    self.settings['Path']['IMOD mrc2tif'],
                    root_name,
                    new_name
                    )
            elif extension == '.tiff' or \
                    extension == '.tif':
                command = 'rsync {0} {1}'.format(root_name, new_name)
            else:
                message = '\n'.join([
                    '{0}: Not known!'.format(self.settings['General']['Output extension']),
                    'Please contact the TranSPHIRE authors.'
                    ])
                self.queue_com['error'].put(message)
                raise IOError(message)

            # Log files
            err_file = os.path.join(
                    self.settings['compress_folder'],
                    '{0}.err'.format(new_root_name)
                    )
            log_file = os.path.join(
                    self.settings['compress_folder'],
                    '{0}.log'.format(new_root_name)
                    )

            # Run the command
            with open(log_file, 'w') as out:
                out.write(command)
                with open(err_file, 'w') as err:
                    start_time = ti.time()
                    sp.Popen(command.split(), stdout=out, stderr=err).wait()
                    stop_time = ti.time()
                    out.write('\nTime: {0} sec'.format(stop_time - start_time))

            tus.check_outputs(
                zero_list=[err_file],
                non_zero_list=[log_file, new_name],
                folder=self.settings['compress_folder'],
                command=command
                )


        # Add to queue
        for aim in self.content_settings['aim']:
            *compare, aim_name = aim.split(':')
            var = True
            for typ in compare:
                name = typ.split('!')[-1]
                if typ.startswith('!'):
                    if self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
                else:
                    if not self.settings['Copy'][name] == 'False':
                        continue
                    else:
                        var = False
                        break
            if var:
                self.add_to_queue(aim=aim_name, root_name=new_name)
                self.add_to_queue(aim=aim_name, root_name=log_file)
                self.add_to_queue(aim=aim_name, root_name=err_file)
            else:
                pass

        if self.settings['compress_folder'] in root_name:
            pass
        elif self.settings['Copy']['Delete stack after compression?'] == 'True':
            os.remove(root_name)
        else:
            pass


    def run_copy_extern(self, root_name):
        """
        Copy to Work/Backup/HDD

        root_name - Root name of the file to copy

        Returns:
        None
        """

        mount_folder_name = '{0}_folder'.format(self.typ)
        mount_name = self.settings['Copy'][self.typ]
        sudo = self.settings['Mount'][mount_name]['Need sudo for copy?']
        protocol = self.settings['Mount'][mount_name]['Protocol']
        new_suffix = root_name.replace(
            self.settings['General']['Project directory'],
            ''
            )
        new_suffix = new_suffix.split('/')
        new_prefix = os.path.relpath(self.settings[mount_folder_name]).split('/')

        if sudo == 'True':
            copy_method = self.copy_as_another_user
        else:
            copy_method = self.copy_as_user

        if protocol == 'hdd':
            for hdd_folder in glob.glob(
                    '{0}/*'.format(
                        self.settings[mount_folder_name]
                        )
                    ):
                if os.path.getsize(root_name) > \
                        sh.disk_usage(hdd_folder).free:
                    new_name = None
                    continue
                else:
                    new_name = os.path.join(
                        *new_prefix,
                        os.path.basename(hdd_folder),
                        *new_suffix
                        )
                    break
            if new_name is None:
                raise IOError('No space on HDD left!')
        else:
            new_name = os.path.join(*new_prefix, *new_suffix)

        copy_method(root_name, new_name)

    def copy_as_user(self, file_in, file_out):
        """
        Copy to another device.

        file_in - Input file path
        file_out - Output file path

        Returns:
        None
        """
        self.check_ready_for_copy(file_out=file_out)

        tu.mkdir_p(os.path.dirname(file_out))
        tu.copy(file_in, file_out)


    def check_ready_for_copy(self, file_out):
        """
        Check, if the current file is ready to copy.

        file_out - File to copy

        Returns:
        True, if ready
        """
        if 'Translation_file.txt' in file_out:
            while True:
                is_locked = bool(not self.shared_dict['translate_lock'].tryLock())
                if not is_locked:
                    self.shared_dict['translate_lock'].unlock()
                    break
                else:
                    QThread.msleep(100)
        elif '_transphire_partres.txt' in file_out:
            while True:
                is_locked = bool(not self.shared_dict['ctf_partres_lock'].tryLock())
                if not is_locked:
                    self.shared_dict['ctf_partres_lock'].unlock()
                    break
                else:
                    QThread.msleep(100)
        elif '_transphire.star' in file_out:
            while True:
                is_locked = bool(not self.shared_dict['ctf_star_lock'].tryLock())
                if not is_locked:
                    self.shared_dict['ctf_star_lock'].unlock()
                    break
                else:
                    QThread.msleep(100)
        else:
            pass
        return True

    def copy_as_another_user(self, file_in, file_out):
        """
        Copy to device as another user via sudo.

        file_in - Input file path
        file_out - Output file path

        Returns:
        None
        """
        self.shared_dict['global_lock'].lock()
        self.check_ready_for_copy(file_out=file_out)

        try:
            self.mkdir_p_as_another_user(folder=os.path.dirname(file_out))
            command = 'sudo -k -S -u {0} rsync {1} {2}'.format(
                self.user,
                file_in,
                file_out
                )
            child = pe.spawnu(command)
            child.sendline(self.password)
            child.expect(pe.EOF)
        except Exception:
            raise
        finally:
            self.shared_dict['global_lock'].unlock()

        if 'rsync:' in child.before:
            text = child.before.replace(self.password, 'PASSWORD')
            raise IOError(
                'Cannot copy file: {0}! {1}'.format(file_out, text)
                )
        else:
            pass

    def mkdir_p_as_another_user(self, folder):
        """
        Create folders recursively as another user with the help of sudo.

        folder - Folder structure to create

        Returns:
        None
        """
        command = 'sudo -k -S -u {0} mkdir -p {1}'.format(self.user, folder)
        child = pe.spawnu(command)
        child.sendline(self.password)
        child.expect(pe.EOF)
        if 'mkdir:' in child.before:
            text = child.before.replace(self.password, 'PASSWORD')
            raise IOError(
                'Cannot create directory: {0}! {1}'.format(folder, text)
                )
        else:
            pass




"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os


def get_motion_default(settings, motion_frames, queue_com, name):
    """
    Set the default values for the motion correction software.

    settings - TranSPHIRE settings.
    motion_frames - Sub frame settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    True, if dose weighting will be applied.
    """

    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        motion_frames['last'] = \
            int(settings['General']['Number of frames']) - \
            int(settings[motion_name]['-Trunc'])
        motion_frames['first'] = \
            int(settings[motion_name]['-Throw']) + 1

        return bool(
                settings[motion_name]['-FmDose'] != '0' and
                settings[motion_name]['-PixSize'] != '0' and
                settings[motion_name]['-kV'] != '0'
                )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)


def get_dw_file_name(output_transfer_scratch, file_name, settings, queue_com, name):
    """
    Get the name of the dose weighted file directly after the program finished.

    output_transfer - Name of the folder in the scratch directory.
    file_name - File name of the root_name path.
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    File path of the DW file.
    """
    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        return os.path.join(
            output_transfer_scratch,
            '{0}_DW.mrc'.format(file_name)
            )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)


def get_motion_command(file_input, file_output_scratch, file_log_scratch, settings, queue_com, name):
    """
    Get the command for the selected motion software.

    file_input - Input file for motion correction.
    file_output_scratch - Output file name
    file_log_scratch - Logfile path on the scratch disc
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    Motion command
    """
    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        return create_motion_cor_2_v1_0_0_command(
            motion_name=settings['Copy']['Motion'],
            file_input=file_input,
            file_output=file_output_scratch,
            file_log=file_log_scratch,
            settings=settings,
            queue_com=queue_com,
            name=name
            )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(settings['Copy']['Motion']),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)


def create_motion_cor_2_v1_0_0_command(motion_name, file_input, file_output, file_log, settings, queue_com, name):
    """
    Create the MotionCor2 v1.0.0 command

    file_input - Input file for motion correction.
    file_output - Output filename
    file_log - Logfile name
    settings - TranSPHIRE settings.
    queue_com - Queue for communication.
    name - Name of the process.

    Returns:
    Command for MotionCor2 v1.0.0
    """

    command = []
    # Start the program
    command.append('{0}'.format(settings['Path'][motion_name]))
    # Input Micrograph
    _, extension = os.path.splitext(file_input)
    if extension == '.tiff' or \
            extension == '.tif':
        command.append('-InTiff')
        command.append('{0}'.format(file_input))

    elif extension == '.mrc':
        command.append('-InMrc')
        command.append('{0}'.format(file_input))

    else:
        message = '{0}: Not known!'.format(extension)
        queue_com['error'].put(message, name)
        raise IOError(message)

    # Output micrograph
    command.append('-OutMrc')
    command.append('{0}'.format(file_output))
    # Write the output stack
    command.append('-OutStack')
    command.append('1')
    # Log file
    command.append('-LogFile')
    command.append('{0}'.format(file_log))

    for key in settings[motion_name]:
        if settings[motion_name][key]:
            command.append(key)
            command.append(
                '{0}'.format(settings[motion_name][key])
                )
        else:
            continue

    return ' '.join(command)


def create_sum_movie_command(
        motion_frames, file_input, file_output, file_shift, file_frc,
        settings, queue_com, name
        ):
    """
    Create the SumMovie command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for SumMovie
    """
    command = create_sum_movie_v1_0_2_command(
        motion_frames=motion_frames,
        file_input=file_input,
        file_output=file_output,
        file_shift=file_shift,
        file_frc=file_frc,
        settings=settings,
        queue_com=queue_com,
        name=name
        )
    return command


def create_sum_movie_v1_0_2_command(
        motion_frames, file_input, file_output, file_shift, file_frc,
        settings, queue_com, name
        ):
    """
    Create the SumMovie v1.0.2 command.

    motion_frames - Sub frames settings dictionary
    file_input - File to sum.
    file_output - Output file name
    file_shift - Output shift file name
    file_frc - Output frc file name
    settings - TranSPHIRE settings
    queue_com - Queue for communication
    name - Name of the process

    Returns:
    Command for Summovie v1.0.2
    """
    sum_movie_command = []
    # Input file
    sum_movie_command.append('{0}'.format(file_input))
    # Number of frames

    motion_name = settings['Copy']['Motion']
    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        sum_movie_command.append('{0}'.format(
            int(settings['General']['Number of frames']) -
            int(settings[motion_name]['-Trunc']) -
            int(settings[motion_name]['-Throw'])
            ))

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    # Output sum file
    sum_movie_command.append('{0}'.format(file_output))
    # Shift file
    sum_movie_command.append('{0}'.format(file_shift))
    # FRC file
    sum_movie_command.append('{0}'.format(file_frc))
    # First frame
    sum_movie_command.append('{0}'.format(motion_frames['first']))
    # Last frame
    sum_movie_command.append('{0}'.format(motion_frames['last']))
    # Pixel size

    if motion_name == 'MotionCor2 v1.0.0' or \
            motion_name == 'MotionCor2 v1.0.5' or \
            motion_name == 'MotionCor2 v1.1.0':
        sum_movie_command.append(
            '{0}'.format(settings[motion_name]['-PixSize'])
            )

    else:
        message = '\n'.join([
            '{0}: Motion version not known.'.format(motion_name),
            'Please contact the TranSPHIRE authors!'
            ])
        queue_com['error'].put(
            message,
            name
            )
        raise IOError(message)

    # Dose correction
    sum_movie_command.append('No')

    command = 'echo "{0}" | {1}'.format(
        '\n'.join(sum_movie_command),
        '{0}'.format(settings['Path']['SumMovie v1.0.2'])
        )

    return command



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np


def get_mic_number(array, settings):
    """
    Identify the micrograph number out of the name string.

    Arguments:
    array - Array containing information
    settings - User provided settings

    Return:
    Array of micrograph numbers
    """
    error = False
    if settings['General']['Rename micrographs'] == 'True':
        number_list = []
        prefix = settings['General']['Rename prefix']
        suffix = settings['General']['Rename suffix']
        for entry in array:
            if suffix == '':
                first_part = entry.split('.'.encode('utf-8'))[0]
            else:
                first_part_split = entry.split(suffix.encode('utf-8'))
                first_part = suffix.encode('utf-8').join(first_part_split[:-1])
            if prefix == '':
                number = first_part
            else:
                number = first_part.split(prefix.encode('utf-8'))[-1]
            try:
                number_list.append(int(number))
            except ValueError:
                error = True
                break
    else:
        number_list = np.arange(len(array))

    if error:
        number_list = np.arange(len(array))
    else:
        pass

    return np.array(number_list)


def update_cter_v1_0(data, settings, label):
    """
    Update the plot for CTER v1.0.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values, y_values, label, title = update_ctffind_4_v4_1_8(
        data=data,
        settings=settings,
        label=label
        )
    return x_values, y_values, label, title


def update_gctf_v1_18(data, settings, label):
    """
    Update the plot for Gctf v1.18.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values, y_values, label, title = update_ctffind_4_v4_1_8(
        data=data,
        settings=settings,
        label=label
        )
    return x_values, y_values, label, title


def update_gctf_v1_06(data, settings, label):
    """
    Update the plot for Gctf v1.06.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values, y_values, label, title = update_ctffind_4_v4_1_8(
        data=data,
        settings=settings,
        label=label
        )
    return x_values, y_values, label, title


def update_ctffind_4_v4_1_10(data, settings, label):
    """
    Update the plot for CTFFIND v4.1.10.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values, y_values, label, title = update_ctffind_4_v4_1_8(
        data=data,
        settings=settings,
        label=label
        )
    return x_values, y_values, label, title


def update_ctffind_4_v4_1_8(data, settings, label):
    """
    Update the plot for CTFFIND v4.1.8.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    if label == 'defocus':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['defocus']/10000
        label = 'Defocus / mum'
        title = 'Mean defocus'

    elif label == 'defocus_diff':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['defocus_diff'] / 10000
        label = 'Defocus diff / mum'
        title = 'Defocus diff'

    elif label == 'astigmatism':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['astigmatism']
        label = 'Angle / degree'
        title = 'Azimuth of astigmatism'

    elif label == 'phase_shift':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['phase_shift']
        label = 'Phase shift / degree'
        title = 'Additional phase shift'

    elif label == 'cross_corr':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['cross_corr']
        label = 'Cross correlation'
        title = 'Cross correlation'

    elif label == 'limit':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['limit']
        y_values[y_values == np.inf] = 0
        label = 'Resolution limit / A'
        title = 'Resolution limit'

    else:
        raise Exception('Plotwidget: Do not know what to do :O {0}'.format(label))

    return x_values, y_values, label, title


def update_motion_cor_2_v1_0_0(data, settings, label):
    """
    Update the plot for MotionCor2 v1.0.0.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    if label == 'overall drift':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['overall drift']
        label = 'Drift / A'
        title = 'Overall drift'

    elif label == 'average drift per frame':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['average drift per frame']
        label = 'Drift / A'
        title = 'Average drift per frame'

    elif label == 'first frame drift':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['first frame drift']
        label = 'Drift / A'
        title = 'First frame drift'

    elif label == 'average drift per frame without first':
        x_values = get_mic_number(data['file_name'], settings)
        y_values = data['average drift per frame without first']
        label = 'Drift / A'
        title = 'Average drift per frame without first'

    else:
        print('Plotwidget: Do not know what to do :O', label)
        raise Exception

    return x_values, y_values, label, title


def update_motion_cor_2_v1_1_0(data, settings, label):
    """
    Update the plot for MotionCor2 v1.1.0.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values, y_values, label, title = update_motion_cor_2_v1_0_0(
        data=data,
        settings=settings,
        label=label
        )
    return x_values, y_values, label, title


def update_motion_cor_2_v1_0_5(data, settings, label):
    """
    Update the plot for MotionCor2 v1.0.5.

    Arguments:
    data - Data to plot
    settings - User provided settings
    label - Label of the plot

    Return:
    x values, y values, label, title
    """
    x_values, y_values, label, title = update_motion_cor_2_v1_0_0(
        data=data,
        settings=settings,
        label=label
        )
    return x_values, y_values, label, title



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    from PyQt4.QtGui import QWidget, QHBoxLayout, QVBoxLayout
    from PyQt4.QtCore import pyqtSlot
except ImportError:
    from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout
    from PyQt5.QtCore import pyqtSlot
from transphire.settingswidget import SettingsWidget
from transphire.separator import Separator


class SettingsContainer(QWidget):
    """
    Widget for setting widgets.

    Inherits:
    QWidget
    """

    def __init__(self, content, max_widgets, parent=None, **kwargs):
        """
        Initialise layout of the widget.

        Arguments:
        content - Content for the widget
        max_widgets - Maximum widgets per column
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(SettingsContainer, self).__init__(parent)

        # Layout
        layout_v_global = QVBoxLayout(self)
        layout_v_global.setContentsMargins(0, 0, 0, 0)
        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout_v = None
        self.idx = 0
        self.max_widgets = max_widgets

        # Global content
        self.content = {}
        self.group = {}

        # Add to v global layout
        layout_v_global.addLayout(self.layout, stretch=0)
        layout_v_global.addStretch(1)

        # Add to layout
        for entry in content:
            for widget in entry:
                for key in widget:
                    if self.idx % self.max_widgets == 0:
                        if self.layout_v is not None:
                            self.layout.addWidget(Separator(typ='vertical', color='lightgrey'))
                            self.layout_v.addStretch(1)
                        self.layout_v = QVBoxLayout()
                        self.layout_v.setContentsMargins(0, 0, 0, 0)
                        self.layout.addLayout(self.layout_v)
                    name = widget[key][1]['name']
                    group = widget[key][1]['group']
                    widget = SettingsWidget(content=widget[key], parent=self)
                    if group:
                        group, state = group.split(':')
                        self.group.setdefault(group, [])
                        self.group[group].append([widget, state, name])
                    self.content[name] = widget
                    self.layout_v.addWidget(widget)
                    self.idx += 1

        for key in self.group:
            self.content[key].sig_index_changed.connect(self.change_state)
            self.change_state(name=key)

        self.layout_v.addStretch(1)
        self.layout.addStretch(1)

    @pyqtSlot(str)
    def change_state(self, name):
        """
        Change the state of widgets based on the choice of the corresponding combo box

        name - Name of the group to change status (Emitted by the combo box)

        Returns:
        None
        """
        try:
            for entry in self.group[name]:
                widget = entry[0]
                state = entry[1]
                sub_name = entry[2]
                if not self.content[name].isEnabled():
                    widget.setEnabled(False)
                elif self.content[name].edit.currentText() == state:
                    widget.setEnabled(True)
                else:
                    widget.setEnabled(False)
                self.change_state(name=sub_name)
        except KeyError:
            return None

    def get_settings(self, quiet=False):
        """
        Get the settings as dict

        Arguments:
        quiet - If True, no prints are executed

        Returns:
        Settings as dictionary
        """
        settings = {}
        error = False
        for key in self.content:
            dictionary = self.content[key].get_settings(quiet=quiet)
            if dictionary is None:
                error = True
            else:
                settings.update(dictionary)
            if key == 'Motion' or key == 'CTF':
                new_key = '{0}_entries'.format(key)
                settings[new_key] = self.content[key].get_combo_entries()
            else:
                pass
        if error:
            return None
        else:
            return [settings]

    def set_settings(self, settings):
        """
        Set settings to the widgets

        Arguments:
        settings - Settings to set.

        Returns:
        None
        """
        for key in settings:
            try:
                content = self.content[key]
            except KeyError:
                if key == 'CTF_entries':
                    continue
                elif key == 'Motion_entries':
                    continue
                else:
                    print('Content for {0} is disabled.'.format(key))
                    continue

            try:
                content.set_settings(settings[key])
            except KeyError:
                if key == 'CTF_entries':
                    pass
                elif key == 'Motion_entries':
                    pass
                else:
                    print(
                        'Setting changed: {0}!'.format(key),
                        'Please do provide new values'
                        )

    def enable(self, var, use_all):
        """
        Disable or enable all widgets

        Arguments:
        var - If True, enable the widgets, else disable them
        use_all - If True, all widgets are enabled/disabled

        Returns:
        None
        """
        for key in self.content:
            if use_all:
                self.content[key].setEnabled(var)
            else:
                self.content[key].setEnabled(var)

        if var:
            try:
                self.set_settings(self.get_settings(quiet=True)[0])
            except TypeError:
                pass
        else:
            pass



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import glob
import re
import numpy as np


def get_header(input_file):
    """
    Extract header information from a relion star file.

    Arguments:
    input_file - Input star file

    Return:
    Color string
    """
    with open(input_file, 'r') as read:
        lines = read.readlines()
    header = []
    idx = None
    for idx, line in enumerate(lines):
        line = line.lstrip().rstrip()
        if line.startswith('_rln'):
            name = line.split()[0]
            name = name.replace('_rlnFinalResolution', '_rlnCtfMaxResolution')
            if name == '_rlnMicrographName' or \
                    name == '_rlnCtfImage':
                header.append((name, '|S200'))
            else:
                header.append((name, '<f8'))
        elif line == '' or line == 'data_' or line == 'loop_':
            continue
        else:
            break

    if idx is None:
        raise ValueError
    else:
        return header, idx


def get_dtype_dict():
    """
    Dtype of the data plot array.

    Arguments:
    None

    Return:
    Dtype dict
    """
    dtype = {}
    dtype['ctf'] = [
        ('mic_number', '<f8'),
        ('defocus', '<f8'),
        ('defocus_diff', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|S200')
        ]
    dtype['Gctf v1.06'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|S200')
        ]
    dtype['Gctf v1.18'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|S200')
        ]
    dtype['CTER v1.0'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|S200')
        ]
    dtype['CTFFIND4 v4.1.10'] = [
        ('mic_number', '<f8'),
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|S200')
        ]
    dtype['CTFFIND4 v4.1.8'] = [
        ('mic_number', '<f8'),
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8'),
        ('file_name', '|S200')
        ]
    dtype['motion'] = [
        ('overall drift', '<f8'),
        ('average drift per frame', '<f8'),
        ('first frame drift', '<f8'),
        ('average drift per frame without first', '<f8'),
        ('file_name', '|S200')
        ]
    return dtype


def get_transphire_dict():
    """
    Translate transphire ctf dict into relion star file information.

    Arguments:
    None

    Return:
    Dtype dict
    """
    transphire_dict = {}
    transphire_dict = {
        'defocus_1': '_rlnDefocusU',
        'defocus_2': '_rlnDefocusV',
        'defocus': '_rlnDefocusU',
        'defocus_diff': '_rlnDefocusV',
        'astigmatism': '_rlnDefocusAngle',
        'phase_shift': '_rlnPhaseShift',
        'cross_corr': '_rlnCtfFigureOfMerit',
        'limit': '_rlnCtfMaxResolution',
        'file_name': '_rlnMicrographName',
        }
    return transphire_dict


def get_relion_dict():
    """
    Translate relion star file information to dtype dict.

    Arguments:
    None

    Return:
    Dtype dict
    """
    relion_dict = {}
    relion_dict = {
        '_rlnDefocusU': 'defocus_1',
        '_rlnDefocusV': 'defocus_2',
        '_rlnDefocusAngle': 'astigmatism',
        '_rlnPhaseShift': 'phase_shift',
        '_rlnCtfFigureOfMerit': 'cross_corr',
        '_rlnCtfMaxResolution': 'limit',
        '_rlnMicrographName': 'file_name',
        }
    return relion_dict


def get_dtype_import_dict():
    """
    Dtype of the file to import.

    Arguments:
    None

    Return:
    Dtype dict
    """
    dtype_import = {}
    dtype_import['CTER v1.0'] = [
        ('defocus', '<f8'),
        ('cs', '<f8'),
        ('volt', '<f8'),
        ('apix', '<f8'),
        ('bfac', '<f8'),
        ('amplitude_contrast', '<f8'),
        ('astigmatism_amplitude', '<f8'),
        ('astigmatism_angle', '<f8'),
        ('standard_deviation_defocus', '<f8'),
        ('standard_deviation_amplitude_contrast', '<f8'),
        ('standard_deviation_astigmatism_amplitude', '<f8'),
        ('standard_deviation_astigmatism_angle', '<f8'),
        ('coefficient_of_variation_of_defocus', '<f8'),
        ('coefficient_of_astigmatism_amplitude', '<f8'),
        ('limit_defocus', '<f8'),
        ('limit_defocus_and_astigmatism', '<f8'),
        ('limit_pixel_error', '<f8'),
        ('limit_maximum', '<f8'),
        ('reserved_spot', '<f8'),
        ('const_amplitude_contrast', '<f8'),
        ('phase_shift', '<f8'),
        ('file_name', '|S200'),
        ]
    dtype_import['Gctf v1.06'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8')
        ]
    dtype_import['Gctf v1.18'] = [
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8')
        ]
    dtype_import['CTFFIND4 v4.1.10'] = [
        ('mic_number', '<f8'),
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8')
        ]
    dtype_import['CTFFIND4 v4.1.8'] = [
        ('mic_number', '<f8'),
        ('defocus_1', '<f8'),
        ('defocus_2', '<f8'),
        ('astigmatism', '<f8'),
        ('phase_shift', '<f8'),
        ('cross_corr', '<f8'),
        ('limit', '<f8')
        ]
    dtype_import['MotionCor2 v1.0.0'] = [
        ('frame_number', '<f8'),
        ('shift_x', '<f8'),
        ('shift_y', '<f8')
        ]
    dtype_import['MotionCor2 v1.0.5'] = [
        ('frame_number', '<f8'),
        ('shift_x', '<f8'),
        ('shift_y', '<f8')
        ]
    dtype_import['MotionCor2 v1.1.0'] = [
        ('frame_number', '<f8'),
        ('shift_x', '<f8'),
        ('shift_y', '<f8')
        ]
    return dtype_import


def import_ctffind_v4_1_10(ctf_name, directory_name):
    """
    Import ctf information for CTFFIND v4.1.10.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    return import_ctffind_v4_1_8(ctf_name, directory_name)


def import_ctffind_v4_1_8(ctf_name, directory_name):
    """
    Import ctf information for CTFFIND v4.1.8.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    files = np.array([
        entry for entry in glob.glob(
            '{0}/*.txt'.format(directory_name)
            ) if '_avrot.txt' not in entry
        ], dtype=str)

    useable_files = []
    for name in files:
        try:
            array = np.genfromtxt(
                name,
                dtype=get_dtype_import_dict()[ctf_name]
                )
        except ValueError:
            continue
        else:
            if array.size > 0:
                useable_files.append(name)
            else:
                continue

    data = np.empty(
        len(useable_files),
        dtype=get_dtype_dict()['ctf']
        )
    data_original = np.empty(
        len(useable_files),
        dtype=get_dtype_dict()[ctf_name]
        )
    data = np.atleast_1d(data)
    data_original = np.atleast_1d(data_original)
    data.fill(0)
    data_original.fill(0)

    for idx, name in sorted(enumerate(useable_files)):

        data_name = np.genfromtxt(
            name,
            dtype=get_dtype_import_dict()[ctf_name]
            )
        data[idx]['file_name'] = name
        input_name = None
        with open(name, 'r') as read:
            for line in read.readlines():
                match_re = re.match('# Input file: (.*?) ; Number of micrographs: 1', line)
                if match_re is not None:
                    input_name = match_re.group(1).encode()
                else:
                    pass
        if input_name is None:
            raise IOError(
                'Could not read {0} file name! Please contact the TranSPHIRE authors!'.format(
                    ctf_name
                    )
                )
        else:
            data_original[idx]['file_name'] = input_name

        for entry in data_name.dtype.names:
            data_original[idx][entry] = data_name[entry]
            if entry == 'defocus_1':
                data[idx]['defocus'] = (data_name['defocus_1']+data_name['defocus_2'])/2
            elif entry == 'defocus_2':
                data[idx]['defocus_diff'] = np.abs(
                    data_name['defocus_1']-data_name['defocus_2']
                    )
            elif entry == 'phase_shift':
                data[idx][entry] = np.degrees(data_name[entry])
            else:
                data[idx][entry] = data_name[entry]

    data = np.sort(data, order='file_name')
    return data, data_original


def import_gctf_v1_18(ctf_name, directory_name):
    """
    Import ctf information for Gctf v1.18.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    data, data_original = import_gctf_v1_06(ctf_name=ctf_name, directory_name=directory_name)
    return data, data_original


def import_gctf_v1_06(ctf_name, directory_name):
    """
    Import ctf information for Gctf v1.06.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    files = np.array(
        [
            entry for entry in glob.glob(
                '{0}/*_gctf.star'.format(directory_name)
            )
            ],
        dtype=str
        )

    useable_files = []
    for file_name in files:
        try:
            dtype, max_header = get_header(input_file=file_name)
            data_name = np.genfromtxt(
                file_name,
                dtype=dtype,
                skip_header=max_header
                )
        except ValueError:
            continue
        else:
            if data_name.size > 0:
                useable_files.append(file_name)
            else:
                continue

    data = np.empty(
        len(useable_files),
        dtype=get_dtype_dict()['ctf']
        )
    data_original = np.empty(
        len(useable_files),
        dtype=get_dtype_dict()[ctf_name]
        )
    data = np.atleast_1d(data)
    data_original = np.atleast_1d(data_original)
    data.fill(0)
    data_original.fill(0)

    relion_dict = get_relion_dict()
    for idx, file_name in enumerate(useable_files):
        try:
            dtype, max_header = get_header(input_file=file_name)
        except ValueError:
            print('Could not read header of {0}!'.format(file_name))
            data, data_original =  None, None
            break
        data_name = np.genfromtxt(
            file_name,
            dtype=dtype,
            skip_header=max_header
            )
        for name in data_name.dtype.names:
            try:
                transphire_name = relion_dict[name]
            except KeyError:
                continue

            try:
                data_original[idx][transphire_name] = np.nan_to_num(data_name[name])
            except ValueError:
                data_original[idx][transphire_name] = 0

            if transphire_name == 'defocus_1':
                try:
                    data[idx]['defocus'] = (
                        data_name['_rlnDefocusU']+data_name['_rlnDefocusV']
                        ) / 2
                except ValueError:
                    data[idx][transphire_name] = 0
            elif transphire_name == 'defocus_2':
                try:
                    data[idx]['defocus_diff'] = np.abs(
                        data_name['_rlnDefocusU']-data_name['_rlnDefocusV']
                        )
                except ValueError:
                    data[idx][transphire_name] = 0
            else:
                try:
                    data[idx][transphire_name] = np.nan_to_num(data_name[name])
                except ValueError:
                    data[idx][transphire_name] = 0

    return data, data_original


def import_cter_v1_0(ctf_name, directory_name):
    """
    Import ctf information for CTER v1.0.
    Defocus in angstrom, phase shift in degree.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    files = np.array(
        [
            entry for entry in glob.glob('{0}/*/partres.txt'.format(directory_name))
            ],
        dtype=str
        )

    useable_files = []
    for file_name in files:
        try:
            data_name = np.genfromtxt(
                file_name,
                dtype=get_dtype_import_dict()[ctf_name]
                )
        except ValueError:
            continue
        else:
            if data_name.size > 0:
                useable_files.append(file_name)
            else:
                continue

    data = np.empty(
        len(useable_files),
        dtype=get_dtype_dict()['ctf']
        )
    data_original = np.empty(
        len(useable_files),
        dtype=get_dtype_import_dict()[ctf_name]
        )
    data = np.atleast_1d(data)
    data_original = np.atleast_1d(data_original)
    data.fill(0)
    data_original.fill(0)

    for idx, name in sorted(enumerate(useable_files)):
        data_name = np.genfromtxt(
            name,
            dtype=get_dtype_import_dict()[ctf_name]
            )
        for entry in data_name.dtype.names:
            data_original[idx][entry] = data_name[entry]
            if entry == 'defocus':
                data[idx][entry] = data_name[entry] * 10000
            elif entry == 'astigmatism_amplitude':
                data[idx]['defocus_diff'] = data_name[entry] * 10000
            elif entry == 'astigmatism_angle':
                data[idx]['astigmatism'] = 45 - data_name[entry]
            elif entry == 'phase_shift':
                data[idx]['phase_shift'] = data_name[entry]
            elif entry == 'file_name':
                data[idx]['file_name'] = data_name[entry]
            elif entry == 'standard_deviation_defocus':
                data[idx]['cross_corr'] = data_name[entry]
            elif entry == 'limit_defocus_and_astigmatism':
                if data_name[entry] == 0:
                    value = data_name['limit_pixel_error']
                else:
                    value = data_name[entry]

                data[idx]['limit'] = 1 / value
            else:
                continue

    return data, data_original


def import_motion_cor_2_v1_0_0(motion_name, directory_names):
    """
    Import motion information for MotionCor2 v1.0.0.

    Arguments:
    motion_name - Name of motion program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    files = np.array(
        [
            entry
            for directory_name in directory_names
            for entry in glob.glob('{0}/*-Full.log'.format(directory_name))
            ],
        dtype=str
        )

    useable_files = []
    for name in files:
        try:
            array = np.genfromtxt(
                name,
                dtype=get_dtype_import_dict()[motion_name]
                )
        except ValueError:
            continue
        else:
            if array.size > 0:
                useable_files.append(name)
            else:
                continue

    data = np.empty(
        len(useable_files),
        dtype=get_dtype_dict()['motion']
        )
    data = np.atleast_1d(data)
    for idx, name in enumerate(useable_files):
        data_name = np.genfromtxt(
            name,
            dtype=get_dtype_import_dict()[motion_name]
            )
        data[idx]['file_name'] = name
        shift_x = np.array([
            data_name['shift_x'][i+1] - data_name['shift_x'][i] \
            for i in range(0, int(data_name['frame_number'][-1]-1))
            ])
        shift_y = np.array([
            data_name['shift_y'][i+1] - data_name['shift_y'][i] \
            for i in range(0, int(data_name['frame_number'][-1]-1))
            ])
        for entry in data.dtype.names:
            if entry == 'overall drift':
                data[idx][entry] = np.sum(np.sqrt(shift_x**2 + shift_y**2))
            elif entry == 'average drift per frame':
                data[idx][entry] = np.sum(np.sqrt(shift_x**2 + shift_y**2))/len(shift_x)
            elif entry == 'first frame drift':
                data[idx][entry] = np.sqrt(shift_x[0]**2 + shift_y[0]**2)
            elif entry == 'average drift per frame without first':
                data[idx][entry] = np.sum(np.sqrt(shift_x[1:]**2 + shift_y[1:]**2))/len(shift_x)
            else:
                pass

    data = np.sort(data, order='file_name')
    return data


def import_motion_cor_2_v1_0_5(motion_name, directory_names):
    """
    Import motion information for MotionCor2 v1.0.5.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    data = import_motion_cor_2_v1_0_0(motion_name=motion_name, directory_names=directory_names)
    return data


def import_motion_cor_2_v1_1_0(motion_name, directory_names):
    """
    Import motion information for MotionCor2 v1.1.0.

    Arguments:
    ctf_name - Name of ctf program
    directory_name - Name of the directory to search for files

    Return:
    Imported data
    """
    data = import_motion_cor_2_v1_0_0(motion_name=motion_name, directory_names=directory_names)
    return data



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import sys
import os
try:
    QT_VERSION = 4
    from PyQt4.QtGui import (
        QMainWindow,
        QHBoxLayout,
        QVBoxLayout,
        QWidget,
        QFileDialog,
        QInputDialog,
        )
    from PyQt4.QtCore import QThread, pyqtSlot, QCoreApplication, QTimer
except ImportError:
    QT_VERSION = 5
    from PyQt5.QtWidgets import (
        QMainWindow,
        QHBoxLayout,
        QVBoxLayout,
        QWidget,
        QFileDialog,
        QInputDialog,
        )
    from PyQt5.QtCore import QThread, pyqtSlot, QCoreApplication, QTimer

# Objects
from transphire.mountworker import MountWorker
from transphire.processworker import ProcessWorker
from transphire.sudopassworddialog import SudoPasswordDialog
from transphire.plotworker import PlotWorker
from transphire.separator import Separator
from transphire.tabdocker import TabDocker
from transphire.mountcalculator import MountCalculator
from transphire import transphire_utils as tu
from transphire import transphire_import as ti


class MainWindow(QMainWindow):
    """
    MainWindow widget.
    Connects GUI and worker threads.

    Inherits from:
    QMainWindow

    Buttons:
    None

    Signals:
    None
    """

    def __init__(
            self, content_gui, content_pipeline, settings_folder,
            mount_directory, version, parent=None
            ):
        """
        Setup the layout for the widget

        Arguments:
        content_gui - Content used to create the GUI outfit.
        content_pipeline - Content used to start processing threads.
        settings_folder - Name of the folder containing settings.
        mount_directory - Name of the folder containing mount points.
        version - Version of TranSPHIRE.
        parent - Parent widget (default None)

        Return:
        None
        """
        super(MainWindow, self).__init__(parent)

        # Ask for sudo password if needed
        self.password = ''
        need_sudo_password = False
        for content in content_gui:
            if content['name'] == 'Status':
                for entry in content['content']:
                    for widget in entry:
                        for key in widget:
                            if key == 'Mount/umount needs sudo password?':
                                if widget[key][0] == 'True':
                                    need_sudo_password = True
                                else:
                                    pass
                            else:
                                pass
            elif content['name'] == 'Mount':
                for entry in content['content_mount']:
                    for widget in entry:
                        for key in widget:
                            if key == 'Need sudo for copy?':
                                if widget[key][0] == 'True':
                                    need_sudo_password = True
                                else:
                                    pass
                            else:
                                pass
            else:
                pass

        if need_sudo_password:
            dialog = SudoPasswordDialog(self)
            dialog.exec_()
            if not dialog.result():
                QCoreApplication.instance().quit()
                sys.exit()
            else:
                self.password = dialog.password
        else:
            pass

        # Start TranSPHIRE from HOME directory.
        self.path = os.environ['HOME']

        # Window title
        self.setWindowTitle(
            '{0} v{1} - {2} - {3}'.format(
                'TranSPHIRE',
                version,
                self.path,
                os.uname()[1]
                )
            )

        # Initiate contents
        self.central_widget = None
        self.content = None
        self.layout = None

        # Settings folder
        self.settings_folder = settings_folder
        self.mount_directory = mount_directory
        self.temp_save = '{0}/temp_save'.format(settings_folder)

        # Threads
        self.mount_worker = None
        self.process_worker = None
        self.plot_worker = None
        self.mount_calculation_ssh = None
        self.mount_calculation_get = None
        self.mount_calculation_df = None
        self.thread_mount = None
        self.thread_process = None
        self.thread_plot = None
        self.mount_thread_list = None

        # Fill GUI
        self.reset_gui(
            content_gui=content_gui, content_pipeline=content_pipeline
            )

    def start_threads(self, content_pipeline):
        """
        Start threads used in TranSPHIRE.

        Arguments:
        content_pipeline - Content used to start processing threads.

        Return:
        None
        """
        # Stop threads if already started.
        if self.mount_worker is not None:
            self.mount_worker.setParent(None)
        if self.process_worker is not None:
            self.process_worker.setParent(None)
        if self.plot_worker is not None:
            self.plot_worker.setParent(None)
        if self.thread_mount is not None:
            self.thread_mount.quit()
            self.thread_mount.wait()
            self.thread_mount.setParent(None)
        if self.thread_process is not None:
            self.thread_process.quit()
            self.thread_process.wait()
            self.thread_process.setParent(None)
        if self.thread_plot is not None:
            self.thread_plot.quit()
            self.thread_plot.wait()
            self.thread_plot.setParent(None)
        if self.mount_thread_list is not None:
            for setting in self.content['Mount'].get_settings():
                for key in setting:
                    thread = self.mount_thread_list[key]['thread']
                    calculator = self.mount_thread_list[key]['object']
                    calculator.kill_thread = True
                    thread.quit()
                    thread.wait()

        # Create objects used in threads
        self.mount_worker = MountWorker(
            password=self.password,
            settings_folder=self.settings_folder,
            mount_directory=self.mount_directory
            )
        self.process_worker = ProcessWorker(
            password=self.password,
            content_process=content_pipeline,
            mount_directory=self.mount_directory
            )
        self.plot_worker = PlotWorker()

        # Create threads
        self.thread_mount = QThread(self)
        self.thread_process = QThread(self)
        self.thread_plot = QThread(self)

        # Start threads
        self.thread_mount.start()
        self.thread_process.start()
        self.thread_plot.start()

        # Start objects in threads
        self.mount_worker.moveToThread(self.thread_mount)
        self.process_worker.moveToThread(self.thread_process)
        self.plot_worker.moveToThread(self.thread_plot)

    def reset_gui(self, content_gui, content_pipeline, load_file=None):
        """
        Reset the content of the mainwindow.

        Arguments:
        content_gui - Content used to fill the GUI.
        content_pipeline - Content used to start processing threads.
        load_file - Settings file (default None).

        Return:
        None
        """
        # Fill MainWindow
        self.set_central_widget()
        self.set_layout_structure()
        self.start_threads(content_pipeline=content_pipeline)
        postprocess_content = self.fill_content(content_gui=content_gui)
        self.postprocess_content(postprocess_content)

        # Load settings saved in load_file
        if load_file is not None:
            self.load(file_name=load_file)
            os.remove('{0}.txt'.format(load_file))
        elif os.path.exists('{0}.txt'.format(self.temp_save)):
            # Result is True if answer is No
            result = tu.question(
                head='Restore previous session.',
                text='Restore previous session?',
                parent=self
                )
            if result:
                pass
            else:
                self.load(file_name=self.temp_save)
        else:
            pass
        self.mount_worker.sig_load_save.emit()

    def save_temp_settings(self):
        """
        Save the status of the GUI in a temp file.

        Arguments:
        None

        Return:
        True, if saving was succesful.
        """
        if os.path.exists('{0}.txt'.format(self.temp_save)):
            os.remove('{0}.txt'.format(self.temp_save))
        else:
            pass
        value = self.save(file_name=self.temp_save, temp=True)
        return value

    def postprocess_content(self, error_list):
        """
        Do postprocessing of creating GUI content, like connecting signals.

        Arguments:
        error_list - List of errors that occured.

        Return:
        True, if saving was succesful.
        """
        for entry in error_list:
            tu.message(entry)

        self.process_worker.sig_finished.connect(self._finished)
        self.process_worker.sig_plot_ctf.connect(self.plot_worker.calculate_array_ctf)
        self.process_worker.sig_plot_motion.connect(self.plot_worker.calculate_array_motion)
        self.plot_worker.sig_message.connect(lambda msg: tu.message(msg))

        self.mount_thread_list = {}
        for key in self.content['Mount'].content:
            thread = QThread(self)
            thread.start()
            mount_calculator = MountCalculator(name=key)
            mount_calculator.moveToThread(thread)
            self.mount_thread_list[key] = {
                'thread': thread,
                'object': mount_calculator
                }
            mount_calculator.sig_finished.connect(
                self.content['Status'].refresh_quota
                )
            mount_calculator.sig_finished.connect(
                self.abort_finished
                )
            self.mount_worker.sig_calculate_ssh_quota.connect(
                mount_calculator.calculate_ssh_quota
                )
            self.mount_worker.sig_calculate_df_quota.connect(
                mount_calculator.calculate_df_quota
                )
            self.mount_worker.sig_calculate_get_quota.connect(
                mount_calculator.calculate_get_quota
                )
        self.content['Mount'].set_threadlist(thread_list=self.mount_thread_list)

    def abort_finished(self, *args, **kwargs):
        """
        Set the mount worker abort variable to True.

        Arguments:
        None

        Return:
        None
        """
        self.mount_worker.abort_finished = True

    def set_central_widget(self):
        """
        Reset the central widget of the MainWindow.

        Arguments:
        None

        Return:
        None
        """
        if self.central_widget is not None:
            self.central_widget.setParent(None)
        else:
            pass

        self.central_widget = QWidget(self)
        self.central_widget.setObjectName('central')
        self.setCentralWidget(self.central_widget)

    def set_layout_structure(self):
        """
        Setup the layout structure for the central widget.

        Arguments:
        None

        Return:
        None
        """
        # Layout dictionary
        self.layout = {}
        self.layout['h1'] = QHBoxLayout(self.central_widget)
        self.layout['h2'] = QHBoxLayout()
        self.layout['h3'] = QHBoxLayout()
        self.layout['v'] = QVBoxLayout()

        # Layout architecture
        self.layout['h1'].addLayout(self.layout['v'], stretch=1)
        self.layout['v'].addLayout(self.layout['h2'], stretch=0)
        self.layout['v'].addWidget(
            Separator(typ='horizontal', color='grey'), stretch=0
            )
        self.layout['v'].addLayout(self.layout['h3'], stretch=1)

    def fill_content(self, content_gui):
        """
        Fill the layouts of the central widget.

        Arguments:
        content_gui - Content used to create the GUI outfit.

        Return:
        List of errors that occured.
        """
        self.content = {}
        exclude_set = tu.get_exclude_set(content=content_gui)
        error_list = []
        tab_list = []
        for entry in content_gui:
            key = entry['name']

            if key in exclude_set:
                continue
            elif entry['layout'] in exclude_set:
                continue
            elif key == 'Stretch':
                layout = entry['layout']
                self.layout[layout].addStretch(1)
                continue
            elif key == 'Separator':
                layout = entry['layout']
                separator = entry['separator']
                self.layout[layout].addWidget(separator)
                continue
            elif key == 'Path':
                tu.reduce_path_widget(exclude_set=exclude_set, content=entry['content'])
            elif key == 'Copy':
                tu.reduce_copy_entries(exclude_set=exclude_set, content=entry['content'])
            else:
                pass

            layout = entry['layout']
            plot_labels = ''
            plot_name = ''
            try:
                plot_name = layout.replace('Plot ', '')
                plot_labels = ti.get_dtype_dict()[
                    tu.get_function_dict()[plot_name]['typ']
                    ]
            except KeyError:
                pass

            # Create widget
            self.content[key] = entry['widget'](
                mount_worker=self.mount_worker,
                process_worker=self.process_worker,
                plot_worker=self.plot_worker,
                settings_folder=self.settings_folder,
                plot_labels=plot_labels,
                plot_name=plot_name,
                parent=self,
                **entry
                )

            if isinstance(self.content[key], TabDocker):
                tab_list.append(key)
            else:
                pass

            if layout in tab_list:
                self.content[layout].add_tab(self.content[key], key)
            else:
                self.layout[layout].addWidget(
                    self.content[key]
                    )

            if key == 'Button':
                self.content[key].sig_load.connect(self.load)
                self.content[key].sig_save.connect(self.save)
                self.content[key].sig_start.connect(self.start)
                self.content[key].sig_stop.connect(self.stop_dialog)
                self.content[key].sig_check_quota.connect(self.check_quota)
            else:
                pass

            if key == 'Notification':
                self.content[key].update_telegram()
                self.content[key].update_email()
                self.content[key].update()
                self.content[key].sig_stop.connect(self.stop)
                timer = QTimer(self)
                timer.setInterval(20000)
                timer.timeout.connect(self.content[key].get_telegram_messages)
                timer.start()
            else:
                pass

            if key == 'Plot per micrograph' or key == 'Plot histogram':
                self.plot_worker.sig_data.connect(
                    self.content[key].update_figure
                    )
            else:
                pass

        return error_list

    def check_quota(self):
        """
        Check the quota for the project and scratch directory.

        Arguments:
        None

        Return:
        None
        """
        global_settings = self.content['General'].get_settings()
        global_settings = {'General': global_settings[0]}
        self.mount_worker.sig_set_settings.emit(global_settings)

    def load(self, file_name=None):
        """
        Load settings from settings file.

        Arguments:
        file_name - Name of the file (default None)

        Return:
        None
        """
        if file_name is None:
            file_name = QFileDialog.getOpenFileName(
                caption='Load settings',
                directory=self.path,
                options=QFileDialog.DontUseNativeDialog
                )

            if QT_VERSION == 4:
                file_name = file_name
            elif QT_VERSION == 5:
                file_name = file_name[0]
            else:
                raise ImportError('QT version unknown! Please contact the transphire authors!')

            if not file_name:
                return
            else:
                pass

        if file_name.endswith('.txt'):
            pass
        elif os.path.exists(file_name):
            pass
        else:
            file_name = '{0}.txt'.format(file_name)


        settings = []
        with open(file_name, 'r') as read:
            for line in read:
                line = line.replace('\n', '')
                key, *value = line.split('\t')
                if isinstance(value, list):
                    settings.append([key, *value])
                else:
                    settings.append([key, value])

        settings = self.settings_to_dict(settings=settings)
        self.set_settings(settings=settings)

    def set_settings(self, settings):
        """
        Load settings from settings file.

        Arguments:
        settings - Settings as dictionary.

        Return:
        None
        """
        for key in settings:
            if key == 'End':
                continue
            else:
                try:
                    self.content[key].set_settings(settings[key])
                except KeyError:
                    print('Key', key, 'no longer exists')
                    continue

    @staticmethod
    def settings_to_dict(settings):
        """
        Make the settings readable for the widgets set settings method.

        Arguments:
        settings - Settings as dictionary.

        Return:
        None
        """
        settings_dict = {}
        idx = -1
        while idx < len(settings)-1:
            idx += 1
            if settings[idx][0] == '###':
                key = settings[idx][1]
                if key == 'Frames':
                    settings_dict[key] = []
                else:
                    settings_dict[key] = {}
                continue

            if len(settings[idx]) == 1:
                settings[idx].append('')

            if key == 'Frames':
                setting = {}
                setting[settings[idx][0]] = settings[idx][1]
                for i in range(3):
                    idx += 1
                    setting[settings[idx][0]] = settings[idx][1]
                settings_dict[key].append(setting)
            else:
                if key == 'Notification':
                    settings_dict[key].update({settings[idx][0]: [
                        settings[idx][1],
                        settings[idx][2]
                        ]})
                else:
                    settings_dict[key].update({
                        settings[idx][0]: settings[idx][1]
                        })

        return settings_dict

    def save(self, file_name=None, temp=False):
        """
        Save GUI status to file.

        Arguments:
        file_name - File name to save settings to.
        temp - File is a temporary save file.

        Return:
        True, if saving was succesful.
        """
        if file_name is None:
            file_name = QFileDialog.getSaveFileName(
                caption='Save settings',
                directory=self.path,
                options=QFileDialog.DontUseNativeDialog,
                filter="Text files (*.txt)"
                )

            if QT_VERSION == 4:
                file_name = file_name
            elif QT_VERSION == 5:
                file_name = file_name[0]
            else:
                raise ImportError('QT version unknown! Please contact the transphire authors!')

            if not file_name:
                return None
            else:
                pass
        else:
            pass

        if file_name.endswith('.txt'):
            pass
        else:
            file_name = '{0}.txt'.format(file_name)

        # Do not override settings
        if os.path.exists(file_name):
            old_filename = file_name
            for number in range(9999):
                file_name = '{0}_{1}'.format(old_filename, number)
                if os.path.exists(file_name):
                    continue
                else:
                    break
        else:
            pass

        error = False
        with open(file_name, 'w') as write:
            for key in self.content:
                if key == 'Mount':
                    continue
                else:
                    pass
                try:
                    settings = self.content[key].get_settings()
                except AttributeError:
                    continue
                if settings is not None:
                    write.write('###\t{0}\n'.format(key))
                    for entry in settings:
                        for key_entry in entry:
                            write.write(
                                '{0}\t{1}\n'.format(key_entry, entry[key_entry])
                                )
                else:
                    error = True
                    message = 'Setting of {0} not valid!'.format(key)
                    tu.message(message)
                    continue
            write.write('###\tEnd\n')
            write.write('The\tEnd\n')

        message_pass = 'Valid settings saved to {0}'.format(file_name)
        message_error = 'Invalid setting detected! Saveing failed!'
        if error:
            os.remove(file_name)
            tu.message(message_error)
            print(message_error)
            return False
        else:
            if temp:
                pass
            else:
                tu.message(message_pass)
            print(message_pass)
            return True

    @pyqtSlot()
    def start(self):
        """
        Start TranSPHIRE processing.

        Arguments:
        None

        Return:
        None
        """
        self.enable(False)
        settings = {}
        # Load settings to pass them to the working threads
        error_list = []
        skip_list = [
            'Mount',
            'Notification',
            'Path',
            'Frames'
            ]
        for key in self.content:
            try:
                settings_widget = self.content[key].get_settings()
            except AttributeError:
                continue
            else:
                settings[key] = {}

            if settings_widget is None:
                self.enable(True)
                return None
            elif key == 'Frames':
                settings_motion = {}
            else:
                pass

            if key == 'Frames':
                skip_name_list = []
            else:
                skip_name_list = tu.get_function_dict()[key]['allow_empty']

            for entry in settings_widget:
                if key not in skip_list:
                    for name in entry:
                        if not entry[name] and name not in skip_name_list:
                            error_list.append(
                                '{0}:{1} is not allowed to be emtpy!'.format(
                                    key,
                                    name
                                    )
                                )
                        else:
                            pass
                else:
                    pass

            for idx, entry in enumerate(settings_widget):
                if key == 'Frames':
                    settings_motion[idx] = entry
                else:
                    settings[key].update(entry)

            if key == 'Frames':
                settings['motion_frames'] = settings_motion
            else:
                pass

        if error_list:
            tu.message('\n'.join(error_list))
            self.enable(True)
            return None
        else:
            pass

        # Get mount information
        for key in settings['Mount']:
            device_name = key.replace(' ', '_')
            save_file = os.path.join(self.settings_folder, device_name)
            try:
                with open(save_file, 'r') as read:
                    lines = read.readlines()
            except FileNotFoundError:
                continue
            for line in lines:
                name = line.split('\t')[0]
                settings['user_{0}'.format(device_name)] = name
        settings['user_Later'] = None

        # Project folder names
        settings['project_folder'] = os.path.join(
            settings['General']['Project directory'],
            settings['General']['Project name']
            )
        settings['compress_folder'] = os.path.join(
            settings['General']['Project directory'],
            settings['General']['Project name']
            )
        settings['motion_folder'] = os.path.join(
            settings['General']['Project directory'],
            settings['General']['Project name']
            )
        settings['ctf_folder'] = os.path.join(
            settings['General']['Project directory'],
            settings['General']['Project name']
            )
        settings['scratch_folder'] = os.path.join(
            settings['General']['Scratch directory'],
            settings['General']['Project name']
            )
        settings['Copy_hdd_folder'] = self.mount_directory
        settings['Copy_backup_folder'] = self.mount_directory
        settings['Copy_work_folder'] = self.mount_directory
        settings['settings_folder'] = os.path.join(
            settings['project_folder'],
            'settings'
            )
        settings['queue_folder'] = os.path.join(
            settings['project_folder'],
            'queue'
            )
        settings['error_folder'] = os.path.join(
            settings['project_folder'],
            'error'
            )

        # Check for continue mode
        if os.path.exists(settings['project_folder']):
            result = self.continue_dialog(
                text1='Output project folder already exists!',
                text2='Do you really want to continue the old run?\nType: YES!'
                )
            if result:
                result_session = self.continue_dialog(
                    text1='Software metafiles',
                    text2='Software metafiles (Atlas, ...) might be already copied!\n' + \
                        'Do you want to copy them again?\nType: YES!'
                    )
                settings['Copy_software_meta'] = bool(result_session)
            else:
                settings['Copy_software_meta'] = True
        else:
            settings['Copy_software_meta'] = True
            result = True

        # Create project and settings folder
        for name in [
                'project_folder', 'settings_folder',
                'scratch_folder', 'queue_folder', 'error_folder'
                ]:
            try:
                tu.mkdir_p(settings[name])
            except FileNotFoundError:
                tu.message('Project name cannot be empty')
                self.enable(True)
                return None

        # Start or stop procedure
        if result:
            self.content['Button'].start_button.setText('Stop')
            self.plot_worker.settings = settings
            self.process_worker.sig_start.emit(settings)
            self.mount_worker.set_settings(settings=settings)
            self.save(
                file_name=os.path.join(
                    settings['settings_folder'],
                    settings['General']['Project name']
                    )
                )
            self.save_temp_settings()
        else:
            tu.message('Input needs to be "YES!" to work')
            self.enable(True)

    def continue_dialog(self, text1, text2):
        """
        Check if the user wants to run the continue mode.

        Arguments:
        text1 - Dialog window name.
        text2 - Text of the dialog.

        Return:
        True, if the input is YES!
        """
        dialog = QInputDialog(self)
        result = dialog.getText(self, text1, text2)
        return bool(result[0] == 'YES!')

    @pyqtSlot()
    def stop_dialog(self):
        """
        Check if the user really wants to stop the process.

        Arguments:
        None

        Return:
        None
        """
        result = self.continue_dialog(
            text1='Do you really want to stop?',
            text2='Do you really want to stop!\nType: YES!'
            )
        if result:
            self.stop()
        else:
            tu.message('Input needs to be "YES!" to work')

    @pyqtSlot()
    def stop(self):
        """
        Stop the process.

        Arguments:
        None

        Return:
        None
        """
        self.process_worker.stop = True
        self.content['Button'].start_button.setEnabled(False)

    @pyqtSlot()
    def _finished(self):
        """
        Rename the Stop Button to start and enable everything.

        Arguments:
        None

        Return:
        None
        """
        self.enable(True)
        self.content['Button'].start_button.setText('Start')
        self.content['Button'].start_button.setEnabled(True)

    @pyqtSlot(bool)
    def enable(self, var, use_all=False):
        """Enable or disable widgets

        Arguments:
        var - Enable status of the widgets.
        use_all - Disable/Enable everything (Default False)

        Return:
        None
        """
        for key in self.content:
            try:
                self.content[key].enable(var=var, use_all=use_all)
            except AttributeError:
                continue

    def closeEvent(self, event):
        """
        Quit threads before close and check if the process is still running

        Arguments:
        event - QCloseEvent.

        Return:
        None
        """
        if self.content['Button'].start_button.text() == 'Stop':
            event.ignore()
            tu.message('First stop the program before closing')
            return None
        elif not self.save_temp_settings():
            result = tu.question(
                head='Error saving file!',
                text='Wrong setting detected! Quit without saving?',
                )
            # Result is true if the answer is No
            if not result:
                pass
            else:
                event.ignore()
                return None
        else:
            pass

        self.thread_mount.quit()
        self.thread_mount.wait()
        self.thread_process.quit()
        self.thread_process.wait()
        self.thread_plot.quit()
        self.thread_plot.wait()
        for key in self.content['Mount'].content:
            thread = self.mount_thread_list[key]['thread']
            calculator = self.mount_thread_list[key]['object']
            calculator.kill_thread = True
            thread.quit()
            thread.wait()
        message = 'Bye Bye'
        print(message)
        super(MainWindow, self).closeEvent(event)



import pkg_resources
from pkg_resources import get_distribution
import os

try:
    _dist = get_distribution('transphire')
    # Normalize case for Windows systems
    dist_loc = os.path.normcase(_dist.location)
    here = os.path.normcase(__file__)
    if not here.startswith(os.path.join(dist_loc, 'transphire')):
        # not installed, but there is another version that *is*
        raise pkg_resources.DistributionNotFound
except pkg_resources.DistributionNotFound:
    __version__ = 'Please install this project with setup.py'
else:
    __version__ = _dist.version



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    from PyQt4.QtGui import QWidget, QCheckBox, QHBoxLayout, QComboBox, QLineEdit
    from PyQt4.QtCore import pyqtSlot
except ImportError:
    from PyQt5.QtWidgets import QWidget, QCheckBox, QHBoxLayout, QComboBox, QLineEdit
    from PyQt5.QtCore import pyqtSlot


class NotificationWidget(QWidget):
    """Widget for notification phone numbers"""

    def __init__(self, name, default, parent=None):
        """
        Initialise layout.

        Arguments:
        name - Name of the widget
        default - Default value for the widget
        parent - Parent widget

        Returns:
        None
        """
        super(NotificationWidget, self).__init__(parent)

        # Global content
        self.name = name
        self.default = default
        if self.default == 'choose':
            self.edit = QComboBox(self)
        else:
            self.edit = QLineEdit(self.name, self)
            self.edit.setReadOnly(True)
        self.check_box = QCheckBox(self.name, self)

        # Event
        self.check_box.stateChanged.connect(self._change_state)

        # Layout
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.check_box)
        layout.addWidget(self.edit)
        self.edit.setEnabled(False)

        # Default
        if self.default == 'choose':
            pass
        else:
            self.edit.setText(default)

        # Object name
        self.edit.setObjectName('noti_edit')
        self.check_box.setObjectName('noti_check')
        self.exceptions = []

    @pyqtSlot()
    def _change_state(self):
        """
        Change enable state of the edit

        Arguments:
        None

        Returns:
        None
        """
        self.edit.setEnabled(self.check_box.isChecked())

    def add_exceptions(self, name):
        """
        Add a person to the exception list and dont send notification anymore.

        Arguments:
        name - Name of the person

        Returns:
        None
        """
        self.exceptions.append(name)

    def update_combo_email(self, users):
        """
        Update the combo boxes.

        Arguments:
        users - User dictionary

        Returns:
        None
        """
        if self.default == 'choose':
            items = ['@ {0}'.format(key) for key in users if '@ {0}'.format(key) not in self.exceptions]
            self.edit.addItems(sorted(items))

    def update_combo_telegram(self, users):
        """
        Update the combo boxes.

        Arguments:
        users - User dictionary

        Returns:
        None
        """
        if self.default == 'choose':
            items = ['T {0}'.format(key) for key in users if 'T {0}'.format(key) not in self.exceptions]
            self.edit.addItems(sorted(items))

    def clear_combo(self):
        """
        Remove all users from the combo box.

        Arguments:
        None

        Returns:
        None
        """
        if self.default == 'choose':
            self.edit.clear()

    def get_settings(self):
        """
        Get text of the currently selected combo item.

        Arguments:
        None

        Returns:
        Settings dictionary
        """
        settings = {}
        if self.default == 'choose':
            settings[self.name] = '{0}\t{1}'.format(
                self.edit.currentText(),
                self.check_box.isChecked()
                )
        else:
            settings[self.name] = '{0}\t{1}'.format(
                self.edit.text(),
                self.check_box.isChecked()
                )
        return settings

    def set_settings(self, name, state):
        """
        Set currently selected combo item in text.

        Arguments:
        name - Name of the person that should be currently selected.
        state - State of the person (True/False;Enables/Disables)

        Returns:
        None
        """
        self.check_box.setChecked(bool(state == 'True'))
        if self.default == 'choose':
            index = self.edit.findText(name)
            self.edit.setCurrentIndex(index)
        else:
            self.edit.setText(name)



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import pexpect as pe
try:
    from PyQt4.QtGui import QInputDialog, QLineEdit
except ImportError:
    from PyQt5.QtWidgets import QInputDialog, QLineEdit
from transphire import transphire_utils as tu


class SudoPasswordDialog(QInputDialog):
    """
    Sudo password dialog.

    Inherits:
    QInputDialog
    """

    def __init__(self, parent=None):
        """
        Initialisation of the SudoPasswordDialog.

        Arguments:
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(SudoPasswordDialog, self).__init__(parent)

        # Global content
        self.password = 'None'

        # Label text
        self.setLabelText('sudo password for {0}:'.format(os.environ['USER']))
        self.setTextEchoMode(QLineEdit.Password)

    def accept(self):
        """
        Modified accept event to check if the sudopassword is valid.

        Arguments:
        None

        Returns:
        None
        """
        self.password = self.textValue()

        child = pe.spawnu('sudo -S -k ls')
        child.sendline(self.password)
        try:
            idx = child.expect([pe.EOF, 'sudo: 1 incorrect password attempt', 'Sorry, try again'], timeout=10)
        except pe.exceptions.TIMEOUT:
            print('Wrong sudo password!')
            tu.message('Wrong sudo password!')
            return None
        else:
            if idx == 0:
                super().accept()
            else:
                print('Wrong sudo password!')
                tu.message('Wrong sudo password!')
                return None



"""
    TranSPHIRE is supposed to help with the cryo-EM data collection
    Copyright (C) 2017 Markus Stabrin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    from PyQt4.QtCore import pyqtSlot, Qt, QEvent
    from PyQt4.QtGui import QMainWindow, QDockWidget, QTabWidget
except ImportError:
    from PyQt5.QtCore import pyqtSlot, Qt, QEvent
    from PyQt5.QtWidgets import QMainWindow, QDockWidget, QTabWidget
from transphire.plotwidget import PlotWidget


class PlotContainer(QMainWindow):
    """
    Qwidget in combination with a FigureCanvas.

    Inherits:
    QMainWindow
    """

    def __init__(self, content, plot_labels, plot_name, *args, parent=None, **kwargs):
        """
        Initialisation of the PlotContainer widget.

        content - Content for the plotcontainer
        plot_lables - Labels of the plot widget
        plot_name - Name of the associated software
        parent - Parent widget (default None)

        Returns:
        None
        """
        super(PlotContainer, self).__init__(parent)
        self.setCentralWidget(None)
        self.setTabPosition(Qt.TopDockWidgetArea, QTabWidget.North)
        self.plot_name = plot_name

        self.content = []
        dock_widgets = []
        for label in plot_labels:
            label = label[0]
            if label == 'mic_number':
                continue
            elif label == 'file_name':
                continue
            else:
                pass

            widget = PlotWidget(label=label, plot_typ=content, parent=self)
            self.content.append(widget)

            dock_widget = QDockWidget(label, self)
            dock_widget.setWidget(widget)
            dock_widget.installEventFilter(self)
            dock_widget.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
            dock_widgets.append(dock_widget)
            self.addDockWidget(Qt.TopDockWidgetArea, dock_widget, Qt.Horizontal)

        for idx in range(1, len(dock_widgets)):
            self.tabifyDockWidget(dock_widgets[0], dock_widgets[idx])

    @pyqtSlot(str, object, str, object)
    def update_figure(self, name, data, directory_name, settings):
        """
        Update the figure in the canvas

        name - Name of the program that called this function.
        data - Data to plot.
        directory_name - Directory to save plots to.
        settings - TranSPHIRE settings

        Returns:
        None
        """
        if self.plot_name == name:
            for entry in self.content:
                entry.update_figure(
                    name=name,
                    data=data,
                    directory_name=directory_name,
                    settings=settings
                    )
        else:
            pass

    def eventFilter(self, source, event):
        """
        Override the QMainWindow eventFilter function.

        source - Source that led to the event trigger
        event - Emitted event

        Returns:
        True, if it has been a close event -> Redock widget
        Event, if it is another event
        """
        if event.type() == QEvent.Close and isinstance(source, QDockWidget):
            event.ignore()
            source.setFloating(False)
            return True
        else:
            return super(PlotContainer, self).eventFilter(source, event)

    def enable(self, var, use_all):
        """
        Enable or disable the widgets.

        var - It True, enable widgets, else disable
        use_all - If True, enable/disable all widgets, else only a subset

        Returns:
        None
        """
        for entry in self.content:
            if use_all:
                entry.setEnabled(var)
            else:
                pass
